"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.apigatewayv2.create_route
hub.exec.boto3.client.apigatewayv2.delete_route
hub.exec.boto3.client.apigatewayv2.get_apis
hub.exec.boto3.client.apigatewayv2.get_route
hub.exec.boto3.client.apigatewayv2.get_routes
hub.exec.boto3.client.apigatewayv2.tag_resource
hub.exec.boto3.client.apigatewayv2.untag_resource
hub.exec.boto3.client.apigatewayv2.update_route
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    api_id: str,
    route_key: str,
    resource_id: str = None,
    api_key_required: bool = None,
    authorization_scopes: List[str] = None,
    authorization_type: str = None,
    authorizer_id: str = None,
    model_selection_expression: str = None,
    operation_name: str = None,
    request_models: Dict = None,
    request_parameters: Dict[
        str,
        make_dataclass("RouteParameters", [("Required", bool, field(default=None))]),
    ] = None,
    route_response_selection_expression: str = None,
    target: str = None,
) -> Dict[str, Any]:
    r"""
    Creates an API Gateway v2 route resource.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        route_key(string): The route key for the route.
        resource_id(string, optional): The route resource identifier in Amazon Web Services.
        api_key_required(bool, optional): Specifies whether an API key is required for the route. Supported only for WebSocket APIs.
        authorization_scopes(List, optional): The authorization scopes supported by this route.
        authorization_type(string, optional): The authorization type for the route. For WebSocket APIs, valid values are NONE for open access,
            AWS_IAM for using AWS IAM permissions, and CUSTOM for using a Lambda authorizer For HTTP APIs,
            valid values are NONE for open access, JWT for using JSON Web Tokens, AWS_IAM for using AWS IAM
            permissions, and CUSTOM for using a Lambda authorizer.
        authorizer_id(string, optional): The identifier of the Authorizer resource to be associated with this route. The authorizer
            identifier is generated by API Gateway when you created the authorizer.
        model_selection_expression(string, optional): The model selection expression for the route. Supported only for WebSocket APIs.
        operation_name(string, optional): The operation name for the route.
        request_models(Dict, optional): The request models for the route. Supported only for WebSocket APIs.
        request_parameters(Dict[str, Dict], optional): The request parameters for the route. Supported only for WebSocket APIs. Defaults to None.
            * (Required, optional): Whether or not the parameter is required.
        route_response_selection_expression(string, optional): The route response selection expression for the route. Supported only for WebSocket APIs.
        target(string, optional): The target for the route.

    Request Syntax:
        [idem_test_aws_apigatewayv2_route]:
          aws.apigatewayv2.route.present:
            - name: 'string'
            - api_id: 'string'
            - route_key: 'string'
            - api_key_required: True|False
            - authorization_scopes: ['string']
            - authorization_type: 'NONE'|'AWS_IAM'|'CUSTOM'|'JWT'
            - authorizer_id: 'string'
            - model_selection_expression: 'string'
            - operation_name: 'string'
            - request_models: {'string': 'string'}
            - request_parameters: {
                'string': {
                    'Required': True|False
                }
              }
            - route_response_selection_expression: 'string'
            - target: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_route:
              aws.apigatewayv2.route.present:
                - name: value
                - api_id: value
                - route_key: value
                - api_key_required: True|False
                - authorization_scopes: []
                - authorization_type: value
                - authorizer_id: value
                - model_selection_expression: value
                - operation_name: value
                - request_models: {}
                - request_parameters: {}
                - route_response_selection_expression: value
                - target: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None
    resource_updated = False

    if resource_id:
        before_ret = await hub.exec.boto3.client.apigatewayv2.get_route(
            ctx, ApiId=api_id, RouteId=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigatewayv2.route", name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.apigatewayv2.route.convert_raw_route_to_present(
            api_id=api_id, raw_resource=before_ret.get("ret"), idem_resource_name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "ApiKeyRequired": api_key_required,
            "AuthorizationScopes": authorization_scopes,
            "AuthorizationType": authorization_type,
            "AuthorizerId": authorizer_id,
            "ModelSelectionExpression": model_selection_expression,
            "OperationName": operation_name,
            "RequestModels": request_models,
            "RequestParameters": request_parameters,
            "RouteKey": route_key,
            "RouteResponseSelectionExpression": route_response_selection_expression,
            "Target": target,
        }

        update_route_ret = await hub.exec.aws.apigatewayv2.route.update(
            ctx,
            api_id=api_id,
            resource_id=resource_id,
            raw_resource=before_ret.get("ret"),
            resource_parameters=resource_parameters,
        )
        result["comment"] = result["comment"] + update_route_ret["comment"]
        if not update_route_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_route_ret["ret"])
        if update_route_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_route_ret["ret"])

        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "api_id": api_id,
                    "api_key_required": api_key_required,
                    "authorization_scopes": authorization_scopes,
                    "authorization_type": authorization_type,
                    "authorizer_id": authorizer_id,
                    "model_selection_expression": model_selection_expression,
                    "operation_name": operation_name,
                    "request_models": request_models,
                    "request_parameters": request_parameters,
                    "route_key": route_key,
                    "route_response_selection_expression": route_response_selection_expression,
                    "target": target,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigatewayv2.route", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.apigatewayv2.create_route(
            ctx,
            ApiId=api_id,
            ApiKeyRequired=api_key_required,
            AuthorizationScopes=authorization_scopes,
            AuthorizationType=authorization_type,
            AuthorizerId=authorizer_id,
            ModelSelectionExpression=model_selection_expression,
            OperationName=operation_name,
            RequestModels=request_models,
            RequestParameters=request_parameters,
            RouteKey=route_key,
            RouteResponseSelectionExpression=route_response_selection_expression,
            Target=target,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigatewayv2.route", name=name
        )
        resource_id = create_ret["ret"]["RouteId"]

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.boto3.client.apigatewayv2.get_route(
            ctx, ApiId=api_id, RouteId=resource_id
        )
        if not after_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + after_ret["comment"]
            return result

        result[
            "new_state"
        ] = hub.tool.aws.apigatewayv2.route.convert_raw_route_to_present(
            api_id=api_id, raw_resource=after_ret.get("ret"), idem_resource_name=name
        )

    return result


async def absent(
    hub, ctx, name: str, api_id: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    Deletes an API Gateway v2 route resource.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string, optional): The Route resource identifier in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_route:
              aws.apigatewayv2.route.absent:
                - name: value
                - api_id: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.route", name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.apigatewayv2.get_route(
        ctx, ApiId=api_id, RouteId=resource_id
    )
    if not before_ret["result"]:
        if "NotFoundException" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.apigatewayv2.route", name=name
            )
        return result

    result["old_state"] = hub.tool.aws.apigatewayv2.route.convert_raw_route_to_present(
        api_id=api_id, raw_resource=before_ret["ret"], idem_resource_name=name
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigatewayv2.route", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.apigatewayv2.delete_route(
        ctx, ApiId=api_id, RouteId=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.apigatewayv2.route", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the API Gateway v2 route resources for an AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigatewayv2.route
    """

    result = {}

    get_apis_ret = await hub.exec.boto3.client.apigatewayv2.get_apis(ctx)
    if not get_apis_ret["result"]:
        hub.log.debug(f"Could not get apis {get_apis_ret['comment']}")
        return result

    for api in get_apis_ret["ret"]["Items"]:
        api_id = api.get("ApiId")

        get_routes_ret = await hub.exec.boto3.client.apigatewayv2.get_routes(
            ctx, ApiId=api_id
        )
        if not get_routes_ret["result"]:
            hub.log.debug(
                f"Could not get routes for ApiId '{api_id}': "
                f"{get_routes_ret['comment']}. Describe will skip this api and continue."
            )
            continue

        for route in get_routes_ret["ret"]["Items"]:
            resource_translated = (
                hub.tool.aws.apigatewayv2.route.convert_raw_route_to_present(
                    api_id=api_id,
                    raw_resource=route,
                )
            )

            result[resource_translated["name"]] = {
                "aws.apigatewayv2.route.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }

    return result


async def search(hub, ctx, name, api_id: str, resource_id: str) -> Dict[str, Any]:
    r"""
    Use an un-managed API Gateway v2 route resource as a data-source.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Route resource identifier in Amazon Web Services.

    Request syntax:
        [idem_test_aws_apigatewayv2_route]:
          aws.apigatewayv2.route.search:
            - api_id: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_route:
              aws.apigatewayv2.route.search:
                - api_id: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    search_ret = await hub.exec.boto3.client.apigatewayv2.get_route(
        ctx, ApiId=api_id, RouteId=resource_id
    )
    if not search_ret["result"]:
        result["result"] = False
        if "NotFoundException" not in str(search_ret["comment"][0]):
            result["comment"] = search_ret["result"]
        else:
            result["comment"] = (
                f"Unable to find aws.apigatewayv2.route resource with resource id '{resource_id}'",
            )
        return result

    result["old_state"] = hub.tool.aws.apigatewayv2.route.convert_raw_route_to_present(
        api_id=api_id, raw_resource=search_ret["ret"], idem_resource_name=name
    )
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result



from typing import Dict
from typing import List


def search_filter_syntax_validation(hub, filters: List) -> Dict:
    """
    Validate if the filter syntax is correct. A filter should contain both "name" and "values" parameter.

    Args:
        hub: The redistributed pop central hub.
        filters: filters input from sls file

    Returns:
        Validation result and error message if the validation fails
    """
    result = {"result": True, "comment": ()}
    if filters is not None:
        for fil in filters:
            if "name" not in fil or "values" not in fil:
                result["comment"] = (f"A filter requires 'name' and 'values'",)
                result["result"] = False
                return result
    return result


def convert_search_filter_to_boto3(hub, filters: List) -> List:
    """
    Convert a filters list in the sls format to the boto3 format. Boto3 uses "Name" and "Values" as parameters.

    Args:
        hub: The redistributed pop central hub.
        filters: A filters list in the sls format

    Returns:
        A filters list in the boto3 format
    """
    result = []
    if filters is not None:
        for fil in filters:
            result.append({"Name": fil["name"], "Values": fil["values"]})
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.iam.attach_role_policy
hub.exec.boto3.client.iam.delete_role_policy
hub.exec.boto3.client.iam.detach_role_policy
hub.exec.boto3.client.iam.get_role_policy
hub.exec.boto3.client.iam.list_role_policies
hub.exec.boto3.client.iam.put_role_policy
resource = hub.tool.boto3.resource.create(ctx, "iam", "RolePolicy", name)
hub.tool.boto3.resource.exec(resource, delete, *args, **kwargs)
hub.tool.boto3.resource.exec(resource, put, *args, **kwargs)
"""
import asyncio
import copy
import re
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.iam.role_policy_attachment.absent",
        ],
    },
    "present": {
        "require": ["aws.iam.role.present"],
    },
}


async def present(
    hub,
    ctx,
    name: str,
    role_name: str,
    policy_document: Dict or str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Adds or updates an inline policy document that is embedded in the specified IAM role. When you embed an inline
    policy in a role, the inline policy is used as part of the role's access (permissions) policy. The role's trust
    policy is created at the same time as the role, using aws.iam.role.present. A role can also have a managed policy
    attached to it. To attach a managed policy to a role, use aws.iam.role_policy_attachment.present.
    To create a new managed policy, use aws.iam.policy.present.
    For information about policies, see Managed policies and inline policies in the IAM User Guide.
    For information about the maximum number of inline policies that you can embed with a role,
    see IAM and STS quotas in the IAM User Guide.

    Args:
        name(Text): The name of the AWS IAM policy.
        role_name(Text): The name of the role to associate the policy with. This parameter allows (through its regex pattern)
         a string of characters consisting of upper and lowercase alphanumeric characters with no spaces.
         You can also include any of the following characters: _+=,.@-
        policy_document(Dict or Text): The policy document. You must provide policies in JSON format in IAM.
         However, for CloudFormation templates formatted in YAML, you can provide the policy in JSON or YAML format.
         CloudFormation always converts a YAML policy to JSON format before submitting it to IAM.
        resource_id(Text, Optional): The role name and policy name with a separator '-'. Format: [role_name]-[policy_name]

    Request Syntax:
        [iam-role-policy-name]:
          aws.iam.role_policy.present:
          - resource_id: 'string'
          - role_name: 'string'
          - policy_document: 'dict or string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-role-policy-930323cb-91cf-42a2-ad9b-3f195c776037:
              aws.iam.role_policy.present:
                - role_name: idem-test-role-e9528a79-a327-4a83-9912-c9b90044f1e4
                - resource_id: idem-test-role-e9528a79-a327-4a83-9912-c9b90044f1e4-idem-test-role-policy-930323cb-91cf-42a2-ad9b-3f195c776037
                - policy_document: '{"Version": "2012-10-17", "Statement": {"Effect": "Allow", "Action": "s3:ListBucket", "Resource": "arn:aws:s3:::example_bucket"}}'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    policy_name = name
    if resource_id:
        if not re.search(f"^({role_name})-({policy_name})$", resource_id):
            result[
                "comment"
            ] = f"Incorrect aws.iam.role_policy resource_id: {resource_id}. Expected id {role_name}-{policy_name}"
            result["result"] = False
            return result
    resource = hub.tool.boto3.resource.create(
        ctx,
        "iam",
        "RolePolicy",
        role_name=role_name,
        name=policy_name,
    )
    before = await hub.tool.boto3.resource.describe(resource)

    # Standardise on the json format
    policy_document = hub.tool.aws.state_comparison_utils.standardise_json(
        policy_document
    )

    if before:
        result[
            "old_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_to_present(before)
        result["new_state"] = copy.deepcopy(result["old_state"])
        result["comment"] = (
            f"aws.iam.role_policy '{policy_name}' for role '{role_name}' already exists",
        )

        if (
            policy_document
            and not hub.tool.aws.state_comparison_utils.is_json_identical(
                result["old_state"]["policy_document"], policy_document
            )
        ):
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    "Would update aws.iam.role_policy policy_document",
                )
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["new_state"]["policy_document"] = policy_document
                return result
            update_ret = await hub.exec.boto3.client.iam.put_role_policy(
                ctx,
                RoleName=role_name,
                PolicyName=name,
                PolicyDocument=policy_document,
            )
            if update_ret["result"]:
                result["comment"] = result["comment"] + (
                    "Updated aws.iam.role_policy policy_document",
                )
            else:
                result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "role_name": role_name,
                        "resource_id": f"{role_name}-{name}",
                        "policy_document": policy_document,
                    },
                )
                result["comment"] = (
                    f"Would create aws.iam.role_policy '{policy_name}' for role '{role_name}'",
                )
                return result

            ret = await hub.exec.boto3.client.iam.put_role_policy(
                ctx,
                RoleName=role_name,
                PolicyName=name,
                PolicyDocument=policy_document,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = (f"Created '{policy_name}' for role  '{role_name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        after = await hub.tool.boto3.resource.describe(resource)
        result[
            "new_state"
        ] = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_to_present(after)
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, role_name: str = None, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified inline policy that is embedded in the specified IAM role. A role can also have managed
    policies attached to it. To detach a managed policy from a role, use aws.iam.role_policy_attachment.absent.
    For more information about policies, refer to Managed policies and inline policies in the IAM User Guide.

    Args:
        name(Text): The name of the AWS IAM policy.
        role_name(Text, optional): The name of the AWS IAM role. Idem automatically considers this resource being absent
         if this field is not specified.
        resource_id(Text, optional): The role name and policy name with a separator '-'. Format: [role_name]-[policy_name].
         If not specified, Idem will use "name" parameter to identify the role policy on AWS.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-test-role-policy-930323cb-91cf-42a2-ad9b-3f195c776037:
              aws.iam.role_policy.absent:
                - name: idem-test-role-policy-930323cb-91cf-42a2-ad9b-3f195c776037
                - role_name: idem-test-role-e9528a79-a327-4a83-9912-c9b90044f1e4
                - resource_id: idem-test-role-e9528a79-a327-4a83-9912-c9b90044f1e4-idem-test-role-policy-930323cb-91cf-42a2-ad9b-3f195c776037
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not role_name:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.role_policy", name=name
        )
        return result
    policy_name = name
    if resource_id:
        if not re.search(f"^({role_name})-({policy_name})$", resource_id):
            result[
                "comment"
            ] = f"Incorrect aws.iam.role_policy resource_id: {resource_id}. Expected id {role_name}-{policy_name}"
            result["result"] = False
            return result
    resource = hub.tool.boto3.resource.create(
        ctx,
        "iam",
        "RolePolicy",
        role_name=role_name,
        name=policy_name,
    )
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = (f"'{policy_name}' for role '{role_name}' already absent",)
    else:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_to_present(
                before
            )

            if ctx.get("test", False):
                result["comment"] = (f"Would delete aws.iam.role_policy {resource_id}",)
                return result

            ret = await hub.exec.boto3.client.iam.delete_role_policy(
                ctx,
                RoleName=role_name,
                PolicyName=policy_name,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = (ret["comment"],)
                result["result"] = False
                return result
            result["comment"] = (f"Deleted '{policy_name}' for role '{role_name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the names of the inline policies that are embedded in of all IAM roles. An IAM role can also have
    managed policies attached to it. These managed polices are not listed with this describe function.
    To list the managed policies that are attached to a role, use aws.iam.role_policy_attachment.describe.
    If there are no inline policies embedded with the specified role, the operation returns an empty dict.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.role_policy
    """

    result = {}
    # To describe all the role policies of all the roles, we first need to list all the roles, then get all the
    # policy names of each role, then get each policy
    ret_roles = await hub.exec.boto3.client.iam.list_roles(ctx)
    if not ret_roles["result"]:
        hub.log.debug(f"Could not describe role {ret_roles['comment']}")
        return {}
    for role in ret_roles["ret"]["Roles"]:
        role_name = role.get("RoleName")
        # List all role policy names of each role
        try:
            ret_role_policies = await hub.exec.boto3.client.iam.list_role_policies(
                ctx=ctx, RoleName=role_name
            )
            if not ret_role_policies["result"]:
                hub.log.warning(
                    f"Failed on fetching role policies with role {role_name} "
                    f"with error {ret_role_policies['comment']}. Describe will skip this role and continue."
                )
                continue
        except hub.tool.boto3.exception.ClientError as e:
            hub.log.warning(
                f"Failed on fetching role policies with role {role_name} with error"
                f" {e.__class__.__name__}: {e} Describe will skip this role and continue."
            )
            continue
        if ret_role_policies:
            if not ret_role_policies["result"]:
                hub.log.warning(
                    f"Could not describe role_policy with role {role_name} with error "
                    f"{ret_role_policies['comment']}. Describe will skip this role and continue."
                )
            else:
                # Get each policy according to the role name and role policy name
                for get_role_policy in asyncio.as_completed(
                    [
                        hub.exec.boto3.client.iam.get_role_policy(
                            ctx=ctx, RoleName=role_name, PolicyName=role_policy_name
                        )
                        for role_policy_name in ret_role_policies["ret"].get(
                            "PolicyNames", list()
                        )
                    ]
                ):
                    ret_role_policy = await get_role_policy
                    if not ret_role_policy["result"]:
                        hub.log.warning(
                            f"Could not get a role_policy with role {role_name} with error"
                            f" {ret_role_policy['comment']} . Describe will skip this role policy and continue."
                        )
                    else:
                        resource = ret_role_policy["ret"]
                        translated_resource = hub.tool.aws.iam.conversion_utils.convert_raw_role_policy_to_present(
                            resource
                        )
                        resource_key = (
                            f"iam-role-policy-{translated_resource['resource_id']}"
                        )
                        result[resource_key] = {
                            "aws.iam.role_policy.present": [
                                {parameter_key: parameter_value}
                                for parameter_key, parameter_value in translated_resource.items()
                            ]
                        }

    return result



from typing import Any
from typing import Dict


async def modify(
    hub, name: str, raw_profile: Dict[str, Any], new_profile: Dict[str, Any]
):
    """
    Copy esm from the raw_profile to the new_profile
    """
    if "esm" in raw_profile:
        new_profile["esm"] = raw_profile["esm"]



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.apigatewayv2.create_stage
hub.exec.boto3.client.apigatewayv2.delete_stage
hub.exec.boto3.client.apigatewayv2.get_apis
hub.exec.boto3.client.apigatewayv2.get_stage
hub.exec.boto3.client.apigatewayv2.get_stages
hub.exec.boto3.client.apigatewayv2.tag_resource
hub.exec.boto3.client.apigatewayv2.untag_resource
hub.exec.boto3.client.apigatewayv2.update_stage
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    api_id: str,
    resource_id: str = None,
    access_log_settings: make_dataclass(
        "AccessLogSettings",
        [
            ("DestinationArn", str, field(default=None)),
            ("Format", str, field(default=None)),
        ],
    ) = None,
    auto_deploy: bool = None,
    client_certificate_id: str = None,
    default_route_settings: make_dataclass(
        "RouteSettings",
        [
            ("DataTraceEnabled", bool, field(default=None)),
            ("DetailedMetricsEnabled", bool, field(default=None)),
            ("LoggingLevel", str, field(default=None)),
            ("ThrottlingBurstLimit", int, field(default=None)),
            ("ThrottlingRateLimit", float, field(default=None)),
        ],
    ) = None,
    deployment_id: str = None,
    description: str = None,
    route_settings: Dict[
        str,
        make_dataclass(
            "RouteSettingsMap",
            [
                ("DataTraceEnabled", bool, field(default=None)),
                ("DetailedMetricsEnabled", bool, field(default=None)),
                ("LoggingLevel", str, field(default=None)),
                ("ThrottlingBurstLimit", int, field(default=None)),
                ("ThrottlingRateLimit", float, field(default=None)),
            ],
        ),
    ] = None,
    stage_variables: Dict[str, str] = None,
    tags: Dict = None,
) -> Dict[str, Any]:
    r"""
    Creates an API Gateway v2 stage resource.

    Args:
        name(string): An Idem name of the resource. This is also used as the name of the Stage during resource creation.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string, optional): The stage resource name in Amazon Web Services.
        access_log_settings(Dict, optional): Settings for logging access in this stage. Defaults to None.
            * (DestinationArn, optional): The ARN of the CloudWatch Logs log group to receive access logs.
            * (Format, optional): A single line format of the access logs of data, as specified by selected $context variables.
                The format must include at least $context.requestId.
        auto_deploy(bool, optional): Specifies whether updates to an API automatically trigger a new deployment.
        client_certificate_id(string, optional): The identifier of a client certificate for a Stage. Supported only for WebSocket APIs.
        default_route_settings(Dict, optional): The default route settings for the stage. Defaults to None.
            * (DataTraceEnabled, optional): Specifies whether (true) or not (false) data trace logging is enabled for this route. This
                property affects the log entries pushed to Amazon CloudWatch Logs. Supported only for WebSocket
                APIs.
            * (DetailedMetricsEnabled, optional): Specifies whether detailed metrics are enabled.
            * (LoggingLevel, optional): Specifies the logging level for this route: INFO, ERROR, or OFF. This property affects the log
                entries pushed to Amazon CloudWatch Logs. Supported only for WebSocket APIs.
            * (ThrottlingBurstLimit, optional): Specifies the throttling burst limit.
            * (ThrottlingRateLimit, optional): Specifies the throttling rate limit.
        deployment_id(string, optional): The deployment identifier of the API stage.
        description(string, optional): The description for the API stage.
        route_settings(Dict[str, Dict], optional): Route settings for the stage, by routeKey. Defaults to None.
            * (DataTraceEnabled, optional): Specifies whether (true) or not (false) data trace logging is enabled for this route. This
                property affects the log entries pushed to Amazon CloudWatch Logs. Supported only for WebSocket
                APIs.
            * (DetailedMetricsEnabled, optional): Specifies whether detailed metrics are enabled.
            * (LoggingLevel, optional): Specifies the logging level for this route: INFO, ERROR, or OFF. This property affects the log
                entries pushed to Amazon CloudWatch Logs. Supported only for WebSocket APIs.
            * (ThrottlingBurstLimit, optional): Specifies the throttling burst limit.
            * (ThrottlingRateLimit, optional): Specifies the throttling rate limit.
        stage_variables(Dict, optional): A map that defines the stage variables for a Stage. Variable names can have alphanumeric and
            underscore characters, and the values must match [A-Za-z0-9-._~:/?#&=,]+.
        tags(Dict, optional): The collection of tags. Each tag element is associated with a given resource.

    Request Syntax:
        [idem_test_aws_apigatewayv2_stage]:
          aws.apigatewayv2.stage.present:
            - name: 'string'
            - api_id: 'string'
            - access_log_settings: {
                'DestinationArn': 'string',
                'Format': 'string'
              }
            - auto_deploy: True|False
            - client_certificate_id: 'string'
            - default_route_settings:{
                'DataTraceEnabled': True|False,
                'DetailedMetricsEnabled': True|False,
                'LoggingLevel': 'ERROR'|'INFO'|'OFF',
                'ThrottlingBurstLimit': int,
                'ThrottlingRateLimit': int
            }
            - deployment_id: 'string'
            - description: 'string'
            - route_settings: {
                'string': {
                    'DataTraceEnabled': True|False,
                    'DetailedMetricsEnabled': True|False,
                    'LoggingLevel': 'ERROR'|'INFO'|'OFF',
                    'ThrottlingBurstLimit': int,
                    'ThrottlingRateLimit': int
                }
              }
            - stage_variables: {'string': 'string'}
            - tags: {'string': 'string'}

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_stage:
              aws.apigatewayv2.stage.present:
                - name: value
                - api_id: value
                - access_log_settings: {}
                - auto_deploy: True|False
                - client_certificate_id: value
                - default_route_settings: {}
                - deployment_id: value
                - description: value
                - route_settings: {}
                - stage_variables: {}
                - tags: {}
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None
    resource_updated = False

    if resource_id:
        before_ret = await hub.exec.boto3.client.apigatewayv2.get_stage(
            ctx, ApiId=api_id, StageName=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigatewayv2.stage", name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.apigatewayv2.stage.convert_raw_stage_to_present(
            api_id=api_id, raw_resource=before_ret.get("ret")
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "AccessLogSettings": access_log_settings,
            "AutoDeploy": auto_deploy,
            "ClientCertificateId": client_certificate_id,
            "DefaultRouteSettings": default_route_settings,
            "DeploymentId": deployment_id,
            "Description": description,
            "RouteSettings": route_settings,
            "StageVariables": stage_variables,
        }

        update_stage_ret = await hub.exec.aws.apigatewayv2.stage.update(
            ctx,
            api_id=api_id,
            resource_id=resource_id,
            raw_resource=before_ret.get("ret"),
            resource_parameters=resource_parameters,
        )
        result["comment"] = result["comment"] + update_stage_ret["comment"]
        if not update_stage_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_stage_ret["ret"])
        if update_stage_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_stage_ret["ret"])

        if tags is not None and tags != result["old_state"].get("tags"):
            resource_arn = hub.tool.aws.arn_utils.build_arn(
                partition="aws",
                service="apigateway",
                region=ctx["acct"]["region_name"],
                resource="/apis/" + api_id + "/stages/" + resource_id,
            )
            update_tags_ret = await hub.exec.aws.apigatewayv2.tag.update_tags(
                ctx,
                resource_arn=resource_arn,
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            if not update_tags_ret["result"]:
                result["result"] = False
                return result

            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if update_tags_ret["ret"] and ctx.get("test", False):
                result["new_state"]["tags"] = update_tags_ret["ret"].get("tags")

        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "api_id": api_id,
                    "access_log_settings": access_log_settings,
                    "auto_deploy": auto_deploy,
                    "client_certificate_id": client_certificate_id,
                    "default_route_settings": default_route_settings,
                    "deployment_id": deployment_id,
                    "description": description,
                    "route_settings": route_settings,
                    "stage_variables": stage_variables,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigatewayv2.stage", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.apigatewayv2.create_stage(
            ctx,
            ApiId=api_id,
            AccessLogSettings=access_log_settings,
            AutoDeploy=auto_deploy,
            ClientCertificateId=client_certificate_id,
            DefaultRouteSettings=default_route_settings,
            DeploymentId=deployment_id,
            Description=description,
            RouteSettings=route_settings,
            StageName=name,
            StageVariables=stage_variables,
            Tags=tags,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigatewayv2.stage", name=name
        )
        resource_id = create_ret["ret"]["StageName"]

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.boto3.client.apigatewayv2.get_stage(
            ctx, ApiId=api_id, StageName=resource_id
        )
        if not after_ret["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + after_ret["comment"]
            return result

        result[
            "new_state"
        ] = hub.tool.aws.apigatewayv2.stage.convert_raw_stage_to_present(
            api_id=api_id, raw_resource=after_ret.get("ret")
        )

    return result


async def absent(
    hub, ctx, name: str, api_id: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    Deletes an API Gateway v2 stage resource.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string, optional): The stage resource name in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_stage:
              aws.apigatewayv2.stage.absent:
                - name: value
                - api_id: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.stage", name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.apigatewayv2.get_stage(
        ctx, ApiId=api_id, StageName=resource_id
    )
    if not before_ret["result"]:
        if "NotFoundException" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.apigatewayv2.stage", name=name
            )
        return result

    result["old_state"] = hub.tool.aws.apigatewayv2.stage.convert_raw_stage_to_present(
        api_id=api_id, raw_resource=before_ret["ret"]
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigatewayv2.stage", name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.apigatewayv2.delete_stage(
        ctx, ApiId=api_id, StageName=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.apigatewayv2.stage", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the API Gateway v2 stage resources for an AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigatewayv2.stage
    """

    result = {}

    get_apis_ret = await hub.exec.boto3.client.apigatewayv2.get_apis(ctx)
    if not get_apis_ret["result"]:
        hub.log.debug(f"Could not get apis {get_apis_ret['comment']}")
        return result

    for api in get_apis_ret["ret"]["Items"]:
        api_id = api.get("ApiId")

        get_stages_ret = await hub.exec.boto3.client.apigatewayv2.get_stages(
            ctx, ApiId=api_id
        )
        if not get_stages_ret["result"]:
            hub.log.debug(
                f"Could not get stages for ApiId '{api_id}': "
                f"{get_stages_ret['comment']}. Describe will skip this api and continue."
            )
            continue

        for stage in get_stages_ret["ret"]["Items"]:
            resource_translated = (
                hub.tool.aws.apigatewayv2.stage.convert_raw_stage_to_present(
                    api_id=api_id,
                    raw_resource=stage,
                )
            )

            result[resource_translated["resource_id"]] = {
                "aws.apigatewayv2.stage.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }

    return result


async def search(hub, ctx, name, api_id: str, resource_id: str) -> Dict[str, Any]:
    r"""
    Use an un-managed API Gateway v2 stage resource as a data-source.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Stage resource name in Amazon Web Services.

    Request syntax:
        [idem_test_aws_apigatewayv2_stage]:
          aws.apigatewayv2.stage.search:
            - api_id: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_stage:
              aws.apigatewayv2.stage.search:
                - api_id: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    search_ret = await hub.exec.boto3.client.apigatewayv2.get_stage(
        ctx, ApiId=api_id, StageName=resource_id
    )
    if not search_ret["result"]:
        result["result"] = False
        if "NotFoundException" not in str(search_ret["comment"][0]):
            result["comment"] = search_ret["result"]
        else:
            result["comment"] = (
                f"Unable to find aws.apigatewayv2.stage resource with resource id '{resource_id}'",
            )
        return result

    result["old_state"] = hub.tool.aws.apigatewayv2.stage.convert_raw_stage_to_present(
        api_id=api_id, raw_resource=search_ret["ret"]
    )
    result["new_state"] = copy.deepcopy(result["old_state"])

    return result



async def gather(hub, profiles):
    """
    load profiles from unencrypted AWS credential files

    Example:
    .. code-block:: yaml

        aws:
          profile_name:
            region_name: us-west-1
            endpoint_url: localhost:992
            aws_access_key_id: my_key_id
            aws_secret_access_key: my_key
            aws_session_token: my_token
    """
    sub_profiles = {}

    for profile, ctx in profiles.get("aws", {}).items():
        sub_profiles[profile] = ctx

    return sub_profiles



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

STATE_NAME = "aws.s3.bucket_encryption"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    bucket: str,
    server_side_encryption_configuration: Dict[str, List],
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    Creates an encryption configuration for an S3 bucket resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The name of the S3 bucket in Amazon Web Services. It must be equal to the bucket parameter.
        bucket(string): The name of the S3 bucket in Amazon Web Services.
        server_side_encryption_configuration(Dict): The server-side-encryption configuration for the S3 bucket.
        timeout(Dict, optional): Timeout configuration for S3 bucket encryption configuration.
            * update (string): Timeout configuration for updating the S3 bucket encryption configuration.
                * delay (int, optional): The amount of time in seconds to wait between attempts. Defaults to 4 seconds.
                * max_attempts (int, optional): Maximum attempts of waiting for the update. Defaults to 30 attempts.

    Request Syntax:
        [idem_test_aws_s3_bucket_encryption]:
          aws.s3.bucket_encryption.present:
            - name: 'string'
            - bucket: 'string'
            - server_side_encryption_configuration: {'string': []}

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_s3_bucket_encryption:
              aws.s3.bucket_encryption.present:
                - name: value
                - bucket: value
                - server_side_encryption_configuration:
                    Rules:
                    - ApplyServerSideEncryptionByDefault:
                        SSEAlgorithm: 'AES256'|'aws:kms'
                        KMSMasterKeyID: 'string'
                      BucketKeyEnabled: True|False
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None

    if resource_id:
        if resource_id != bucket:
            result["result"] = False
            result["comment"] = (
                f"Bucket '{bucket}' and resource_id '{resource_id}' parameters must be the same",
            )
            return result

        before_ret = await hub.exec.boto3.client.s3.get_bucket_encryption(
            ctx, Bucket=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type=STATE_NAME, name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
            bucket=resource_id,
            raw_resource=before_ret.get("ret"),
            idem_resource_name=name,
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "server_side_encryption_configuration": server_side_encryption_configuration,
        }

        if hub.tool.aws.s3.bucket_encryption.is_bucket_encryption_configuration_identical(
            resource_parameters, result["old_state"]
        ):
            # no updates to apply
            return result

        if ctx.get("test", False):
            result["new_state"].update(resource_parameters)
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.would_update_comment(
                resource_type=STATE_NAME, name=name
            )
            return result
        else:
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment(
                resource_type=STATE_NAME, name=name
            )
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "bucket": bucket,
                    "server_side_encryption_configuration": server_side_encryption_configuration,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=STATE_NAME, name=name
            )
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type=STATE_NAME, name=name
        )

    put_ret = await hub.exec.boto3.client.s3.put_bucket_encryption(
        ctx,
        Bucket=bucket,
        ServerSideEncryptionConfiguration=server_side_encryption_configuration,
    )
    if not put_ret["result"]:
        result["result"] = False
        result["comment"] = result["comment"] + put_ret["comment"]
        return result

    wait_for_updates_ret = await hub.tool.aws.s3.bucket_encryption.wait_for_updates(
        ctx,
        bucket=bucket,
        server_side_encryption_configuration=server_side_encryption_configuration,
        timeout=timeout.get("update") if timeout else None,
    )
    if not wait_for_updates_ret["result"]:
        result["result"] = False
        result["comment"] += wait_for_updates_ret["comment"]
        return result

    result[
        "new_state"
    ] = hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
        bucket=bucket,
        raw_resource=wait_for_updates_ret.get("ret"),
        idem_resource_name=name,
    )

    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, timeout: Dict = None
) -> Dict[str, Any]:
    r"""
    Deletes an encryption configuration from an S3 bucket resource.

    Args:
        name(string): An Idem name of the resource.
        resource_id(string, optional): The name of the S3 bucket in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.
        timeout(Dict, optional): Timeout configuration for S3 bucket encryption configuration.
            * delete (string): Timeout configuration for deleting the S3 bucket encryption configuration.
                * delay (int, optional): The amount of time in seconds to wait between attempts. Defaults to 4 seconds.
                * max_attempts (int, optional): Maximum attempts of waiting for the deletion. Defaults to 30 attempts.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_s3_bucket_encryption:
              aws.s3.bucket_encryption.absent:
                - name: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=STATE_NAME, name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.s3.get_bucket_encryption(
        ctx, Bucket=resource_id
    )
    if not before_ret["result"]:
        if "ServerSideEncryptionConfigurationNotFoundError" not in str(
            before_ret["comment"][0]
        ):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type=STATE_NAME, name=name
            )
        return result

    result[
        "old_state"
    ] = hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
        bucket=resource_id, raw_resource=before_ret["ret"], idem_resource_name=name
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=STATE_NAME, name=name
        )
        return result

    delete_ret = await hub.exec.boto3.client.s3.delete_bucket_encryption(
        ctx, Bucket=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    wait_for_delete_ret = await hub.tool.aws.s3.bucket_encryption.wait_for_delete(
        ctx,
        bucket=resource_id,
        timeout=timeout.get("delete") if timeout else None,
    )
    if not wait_for_delete_ret["result"]:
        result["result"] = False
        result["comment"] = wait_for_delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type=STATE_NAME, name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the encryption configuration for each S3 bucket under the given AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.s3.bucket_encryption
    """

    result = {}

    list_buckets_ret = await hub.exec.boto3.client.s3.list_buckets(ctx)
    if not list_buckets_ret["result"]:
        hub.log.debug(f"Could not list S3 buckets: {list_buckets_ret['comment']}")
        return result

    for bucket in list_buckets_ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")

        get_bucket_encryption_ret = (
            await hub.exec.boto3.client.s3.get_bucket_encryption(
                ctx, Bucket=bucket_name
            )
        )
        if not get_bucket_encryption_ret["result"]:
            if "ServerSideEncryptionConfigurationNotFoundError" not in str(
                get_bucket_encryption_ret["comment"][0]
            ):
                hub.log.debug(
                    f"Could not get encryption configuration for S3 bucket '{bucket_name}': "
                    f"{get_bucket_encryption_ret['comment']}. Describe will skip this S3 bucket and continue."
                )
            continue

        resource_translated = (
            hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
                bucket=bucket_name,
                raw_resource=get_bucket_encryption_ret["ret"],
            )
        )

        result[resource_translated["name"]] = {
            f"{STATE_NAME}.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.autoscaling.create_auto_scaling_group
hub.exec.boto3.client.autoscaling.delete_auto_scaling_group
hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups
hub.exec.boto3.client.autoscaling.update_auto_scaling_group
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    min_size: int,
    max_size: int,
    resource_id: str = None,
    launch_configuration_name: str = None,
    launch_template: make_dataclass(
        "LaunchTemplateSpecification",
        [
            ("LaunchTemplateId", str, field(default=None)),
            ("LaunchTemplateName", str, field(default=None)),
            ("Version", str, field(default=None)),
        ],
    ) = None,
    mixed_instances_policy: make_dataclass(
        "MixedInstancesPolicy",
        [
            ("LaunchTemplate", Dict, field(default=None)),
            ("InstancesDistribution", Dict, field(default=None)),
        ],
    ) = None,
    instance_id: str = None,
    desired_capacity: int = None,
    default_cooldown: int = None,
    availability_zones: List[str] = None,
    load_balancer_names: List[str] = None,
    target_group_ar_ns: List[str] = None,
    health_check_type: str = None,
    health_check_grace_period: int = None,
    placement_group: str = None,
    vpc_zone_identifier: str = None,
    termination_policies: List[str] = None,
    new_instances_protected_from_scale_in: bool = None,
    capacity_rebalance: bool = None,
    lifecycle_hook_specification_list: List[
        make_dataclass(
            "LifecycleHookSpecification",
            [
                ("LifecycleHookName", str, field(default=None)),
                ("LifecycleTransition", str, field(default=None)),
                ("NotificationMetadata", str, field(default=None)),
                ("HeartbeatTimeout", int, field(default=None)),
                ("DefaultResult", str, field(default=None)),
                ("NotificationTargetARN", str, field(default=None)),
                ("RoleARN", str, field(default=None)),
            ],
        )
    ] = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [
                ("key", str, field(default=None)),
                ("value", str, field(default=None)),
                ("propagate_at_launch", bool, field(default=None)),
            ],
        )
    ] = None,
    service_linked_role_arn: str = None,
    max_instance_lifetime: int = None,
    desired_capacity_type: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

     We strongly recommend using a launch template when calling this operation to ensure full functionality for
    Amazon EC2 Auto Scaling and Amazon EC2.  Creates an Auto Scaling group with the specified name and attributes.
    If you exceed your maximum limit of Auto Scaling groups, the call fails. To query this limit, call the
    DescribeAccountLimits API. For information about updating this limit, see Amazon EC2 Auto Scaling service quotas
    in the Amazon EC2 Auto Scaling User Guide. For introductory exercises for creating an Auto Scaling group, see
    Getting started with Amazon EC2 Auto Scaling and Tutorial: Set up a scaled and load-balanced application in the
    Amazon EC2 Auto Scaling User Guide. For more information, see Auto Scaling groups in the Amazon EC2 Auto Scaling
    User Guide. Every Auto Scaling group has three size parameters (DesiredCapacity, MaxSize, and MinSize). Usually,
    you set these sizes based on a specific number of instances. However, if you configure a mixed instances policy
    that defines weights for the instance types, you must specify these sizes with the same units that you use for
    weighting instances.

    Args:
        resource_id(Text, optional): The name of the Auto Scaling group. This name must be unique per Region per account.
        name(Text): An Idem name of the resource. This is also used as the AutoScaling Group Name during resource creation.
        launch_configuration_name(Text, optional): The name of the launch configuration to use to launch instances.  Conditional: You must specify
            either a launch template (LaunchTemplate or MixedInstancesPolicy) or a launch configuration
            (LaunchConfigurationName or InstanceId). Defaults to None.
        launch_template(Dict, optional): Parameters used to specify the launch template and version to use to launch instances.
            Conditional: You must specify either a launch template (LaunchTemplate or MixedInstancesPolicy)
            or a launch configuration (LaunchConfigurationName or InstanceId).  The launch template that is
            specified must be configured for use with an Auto Scaling group. For more information, see
            Creating a launch template for an Auto Scaling group in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
            * (LaunchTemplateId, optional): The ID of the launch template. To get the template ID, use the Amazon EC2
                DescribeLaunchTemplates API operation. New launch templates can be created using the Amazon EC2
                CreateLaunchTemplate API.  Conditional: You must specify either a LaunchTemplateId or a
                LaunchTemplateName.
            * (LaunchTemplateName, optional): The name of the launch template. To get the template name, use the Amazon EC2
                DescribeLaunchTemplates API operation. New launch templates can be created using the Amazon EC2
                CreateLaunchTemplate API.  Conditional: You must specify either a LaunchTemplateId or a
                LaunchTemplateName.
            * (Version, optional): The version number, $Latest, or $Default. To get the version number, use the Amazon EC2
                DescribeLaunchTemplateVersions API operation. New launch template versions can be created using
                the Amazon EC2 CreateLaunchTemplateVersion API. If the value is $Latest, Amazon EC2 Auto Scaling
                selects the latest version of the launch template when launching instances. If the value is
                $Default, Amazon EC2 Auto Scaling selects the default version of the launch template when
                launching instances. The default value is $Default.
        mixed_instances_policy(Dict, optional): An embedded object that specifies a mixed instances policy. For more information, see Auto
            Scaling groups with multiple instance types and purchase options in the Amazon EC2 Auto Scaling
            User Guide. Defaults to None.
            * (LaunchTemplate, optional): Specifies the launch template to use and the instance types (overrides) that are used to launch
                EC2 instances to fulfill On-Demand and Spot capacities. Required when creating a mixed instances
                policy.
            * (InstancesDistribution, optional): Specifies the instances distribution.
        instance_id(Text, optional): The ID of the instance used to base the launch configuration on. If specified, Amazon EC2 Auto
            Scaling uses the configuration values from the specified instance to create a new launch
            configuration. To get the instance ID, use the Amazon EC2 DescribeInstances API operation. For
            more information, see Creating an Auto Scaling group using an EC2 instance in the Amazon EC2
            Auto Scaling User Guide. Defaults to None.
        min_size(int): The minimum size of the group.
        max_size(int): The maximum size of the group.  With a mixed instances policy that uses instance weighting,
            Amazon EC2 Auto Scaling may need to go above MaxSize to meet your capacity requirements. In this
            event, Amazon EC2 Auto Scaling will never go above MaxSize by more than your largest instance
            weight (weights that define how many units each instance contributes to the desired capacity of
            the group).
        desired_capacity(int, optional): The desired capacity is the initial capacity of the Auto Scaling group at the time of its
            creation and the capacity it attempts to maintain. It can scale beyond this capacity if you
            configure auto scaling. This number must be greater than or equal to the minimum size of the
            group and less than or equal to the maximum size of the group. If you do not specify a desired
            capacity, the default is the minimum size of the group. Defaults to None.
        default_cooldown(int, optional): The amount of time, in seconds, after a scaling activity completes before another scaling
            activity can start. The default value is 300. This setting applies when using simple scaling
            policies, but not when using other scaling policies or scheduled scaling. For more information,
            see Scaling cooldowns for Amazon EC2 Auto Scaling in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
        availability_zones(List, optional): A list of Availability Zones where instances in the Auto Scaling group can be created. This
            parameter is optional if you specify one or more subnets for VPCZoneIdentifier. Conditional: If
            your account supports EC2-Classic and VPC, this parameter is required to launch instances into
            EC2-Classic. Defaults to None.
        load_balancer_names(List, optional): A list of Classic Load Balancers associated with this Auto Scaling group. For Application Load
            Balancers, Network Load Balancers, and Gateway Load Balancers, specify the TargetGroupARNs
            property instead. Defaults to None.
        target_group_ar_ns(List, optional): The Amazon Resource Names (ARN) of the target groups to associate with the Auto Scaling group.
            Instances are registered as targets in a target group, and traffic is routed to the target
            group. For more information, see Elastic Load Balancing and Amazon EC2 Auto Scaling in the
            Amazon EC2 Auto Scaling User Guide. Defaults to None.
        health_check_type(Text, optional): The service to use for the health checks. The valid values are EC2 (default) and ELB. If you
            configure an Auto Scaling group to use load balancer (ELB) health checks, it considers the
            instance unhealthy if it fails either the EC2 status checks or the load balancer health checks.
            For more information, see Health checks for Auto Scaling instances in the Amazon EC2 Auto
            Scaling User Guide. Defaults to None.
        health_check_grace_period(int, optional): The amount of time, in seconds, that Amazon EC2 Auto Scaling waits before checking the health
            status of an EC2 instance that has come into service and marking it unhealthy due to a failed
            health check. The default value is 0. For more information, see Health check grace period in the
            Amazon EC2 Auto Scaling User Guide. Required if you are adding an ELB health check. Defaults to None.
        placement_group(Text, optional): The name of an existing placement group into which to launch your instances, if any. A placement
            group is a logical grouping of instances within a single Availability Zone. You cannot specify
            multiple Availability Zones and a placement group. For more information, see Placement Groups in
            the Amazon EC2 User Guide for Linux Instances. Defaults to None.
        vpc_zone_identifier(Text, optional): A comma-separated list of subnet IDs for a virtual private cloud (VPC) where instances in the
            Auto Scaling group can be created. If you specify VPCZoneIdentifier with AvailabilityZones, the
            subnets that you specify for this parameter must reside in those Availability Zones.
            Conditional: If your account supports EC2-Classic and VPC, this parameter is required to launch
            instances into a VPC. Defaults to None.
        termination_policies(List, optional): A policy or a list of policies that are used to select the instance to terminate. These policies
            are executed in the order that you list them. For more information, see Controlling which Auto
            Scaling instances terminate during scale in in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
        new_instances_protected_from_scale_in(bool, optional): Indicates whether newly launched instances are protected from termination by Amazon EC2 Auto
            Scaling when scaling in. For more information about preventing instances from terminating on
            scale in, see Using instance scale-in protection in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
        capacity_rebalance(bool, optional): Indicates whether Capacity Rebalancing is enabled. Otherwise, Capacity Rebalancing is disabled.
            When you turn on Capacity Rebalancing, Amazon EC2 Auto Scaling attempts to launch a Spot
            Instance whenever Amazon EC2 notifies that a Spot Instance is at an elevated risk of
            interruption. After launching a new instance, it then terminates an old instance. For more
            information, see Amazon EC2 Auto Scaling Capacity Rebalancing in the Amazon EC2 Auto Scaling
            User Guide. Defaults to None.
        lifecycle_hook_specification_list(List, optional): One or more lifecycle hooks for the group, which specify actions to perform when Amazon EC2 Auto
            Scaling launches or terminates instances. Defaults to None.
            * (LifecycleHookName): The name of the lifecycle hook.
            * (LifecycleTransition): The state of the EC2 instance to which you want to attach the lifecycle hook.
                The valid values are:
                    autoscaling:EC2_INSTANCE_LAUNCHING
                    autoscaling:EC2_INSTANCE_TERMINATING
            * (NotificationMetadata, optional): Additional information that you want to include any time Amazon EC2 Auto
                Scaling sends a message to the notification target.
            * (HeartbeatTimeout, optional): The maximum time, in seconds, that can elapse before the lifecycle hook times out. If the
                lifecycle hook times out, Amazon EC2 Auto Scaling performs the action that you specified in the
                DefaultResult parameter. You can prevent the lifecycle hook from timing out by calling
                RecordLifecycleActionHeartbeat.
            * (DefaultResult, optional): Defines the action the Auto Scaling group should take when the lifecycle hook timeout elapses or
                if an unexpected failure occurs. The valid values are CONTINUE and ABANDON. The default value is
                ABANDON.
            * (NotificationTargetARN, optional): The ARN of the target that Amazon EC2 Auto Scaling sends notifications to when an instance is in
                the transition state for the lifecycle hook. The notification target can be either an SQS queue
                or an SNS topic.
            * (RoleARN, optional): The ARN of the IAM role that allows the Auto Scaling group to publish to the specified
                notification target. Valid only if the notification target is an Amazon SNS topic or an Amazon
                SQS queue. Required for new lifecycle hooks, but optional when updating existing hooks.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value, propagate_at_launch-{tag-key}: value}
            or list of tags in the format of [{"key": tag-key, "value": tag-value, "propagate_at_launch": value}].
            One or more tags. You can tag your Auto Scaling group and propagate the tags to the Amazon EC2
            instances it launches. Tags are not propagated to Amazon EBS volumes. To add tags to Amazon EBS
            volumes, specify the tags in a launch template but use caution. If the launch template specifies
            an instance tag with a key that is also specified for the Auto Scaling group, Amazon EC2 Auto
            Scaling overrides the value of that instance tag with the value specified by the Auto Scaling
            group. For more information, see Tagging Auto Scaling groups and instances in the Amazon EC2
            Auto Scaling User Guide. Defaults to None.
            * (key): The tag key.
            * (value, optional): The tag value.
            * (propagate_at_launch, optional): Determines whether the tag is added to new instances as they are launched in the group.
        service_linked_role_arn(Text, optional): The Amazon Resource Name (ARN) of the service-linked role that the Auto Scaling group uses to
            call other Amazon Web Services on your behalf. By default, Amazon EC2 Auto Scaling uses a
            service-linked role named AWSServiceRoleForAutoScaling, which it creates if it does not exist.
            For more information, see Service-linked roles in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
        max_instance_lifetime(int, optional): The maximum amount of time, in seconds, that an instance can be in service. The default is null.
            If specified, the value must be either 0 or a number equal to or greater than 86,400 seconds (1
            day). For more information, see Replacing Auto Scaling instances based on maximum instance
            lifetime in the Amazon EC2 Auto Scaling User Guide. Defaults to None.
        context(Text, optional): Reserved. Defaults to None.
        desired_capacity_type(Text, optional): The unit of measurement for the value specified for desired capacity. Amazon EC2 Auto Scaling
            supports DesiredCapacityType for attribute-based instance type selection only. For more
            information, see Creating an Auto Scaling group using attribute-based instance type selection in
            the Amazon EC2 Auto Scaling User Guide. By default, Amazon EC2 Auto Scaling specifies units,
            which translates into number of instances. Valid values: units | vcpu | memory-mib. Defaults to None.

    Request Syntax:
       [auto-scaling-group-resource-id]:
            aws.autoscaling.auto_scaling_group.present:
              - resource_id: 'string'
              - name: 'string'
              - launch_configuration_name: 'string'
              - min_size: 'int'
              - max_size: 'int'
              - desired_capacity: 'int'
              - default_cooldown: 'int'
              - availability_zones: List
              - health_check_type: 'string'
              - health_check_grace_period: 'int'
              - vpc_zone_identifier: 'string'
              - termination_policies: List
              - new_instances_protected_from_scale_in: 'bool'
              - tags: Dict or List


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-scaling-group:
              aws.autoscaling.auto_scaling_group.present:
              - resource_id: idem-scaling-group
              - name: idem-scaling-group
              - launch_configuration_name: idem-launch_configuration
              - min_size: 2
              - max_size: 4
              - desired_capacity: 2
              - default_cooldown: 300
              - availability_zones:
                - us-west-1b
                - us-west-1c
              - health_check_type: EC2
              - health_check_grace_period: 300
              - vpc_zone_identifier: subnet-xxxxxxx,subnet-xxxxxxx,subnet-xxxxxx
              - termination_policies:
                - Default
              - new_instances_protected_from_scale_in: false
              - tags(tags in list format):
                - key: Identifier
                  value: idem-aws
                  propagate_at_launch: true
                - key: Name
                  propagate_at_launch: false
                  value: idem-scaling-group
              - tags(or in dict format):
                  Identifier: idem-aws
                  propagate_at_launch-Identifier: true
                  Name: idem-scaling-group
                  propagate_at_launch-Name: false
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Query for auto scaling groups
    auto_scaling_group = None
    if resource_id:
        auto_scaling_group = (
            await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(
                ctx, AutoScalingGroupNames=[resource_id]
            )
        )
        if not auto_scaling_group["result"]:
            result["comment"] = auto_scaling_group["comment"]
            result["result"] = auto_scaling_group["result"]
            return result
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_auto_scaling_tag_list_to_dict(tags)

    # Check for existing auto scaling group
    before = None
    if auto_scaling_group and auto_scaling_group["ret"]["AutoScalingGroups"]:
        before = auto_scaling_group["ret"]["AutoScalingGroups"][0]

    # Update current state
    current_state = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
        auto_scaling_group=before
    )
    result["old_state"] = current_state

    # Handle No change behaviour
    desired_state = {
        "name": name,
        "resource_id": resource_id,
        "launch_configuration_name": launch_configuration_name,
        "launch_template": launch_template,
        "instance_id": instance_id,
        "min_size": min_size,
        "max_size": max_size,
        "desired_capacity": desired_capacity,
        "default_cooldown": default_cooldown,
        "availability_zones": availability_zones,
        "load_balancer_names": load_balancer_names,
        "target_group_ar_ns": target_group_ar_ns,
        "health_check_type": health_check_type,
        "health_check_grace_period": health_check_grace_period,
        "placement_group": placement_group,
        "mixed_instances_policy": mixed_instances_policy,
        "vpc_zone_identifier": vpc_zone_identifier,
        "termination_policies": termination_policies,
        "new_instances_protected_from_scale_in": new_instances_protected_from_scale_in,
        "capacity_rebalance": capacity_rebalance,
        "lifecycle_hook_specification_list": lifecycle_hook_specification_list,
        "tags": tags,
        "max_instance_lifetime": max_instance_lifetime,
        "desired_capacity_type": desired_capacity_type,
    }

    is_auto_scaling_group_updated = (
        hub.tool.aws.autoscaling.auto_scaling_group_utils.is_auto_scaling_group_updated(
            current_state=current_state, desired_state=desired_state
        )
    )

    is_tags_updated = current_state.get("tags") != desired_state.get("tags")

    is_change_detected = (
        before is None or is_auto_scaling_group_updated or is_tags_updated
    )

    if not is_change_detected:
        result["comment"] = (
            f"aws.autoscaling.auto_scaling_group '{name}' already exists",
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    # Handle test behaviour
    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state=current_state,
            desired_state=desired_state,
        )
        result["comment"] = (
            hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.autoscaling.auto_scaling_group", name=name
            )
            if before
            else hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.autoscaling.auto_scaling_group", name=name
            )
        )
        return result

    # Handle actual resource create or update
    if before:
        # Update auto scaling group
        if is_auto_scaling_group_updated:
            ret = await hub.exec.boto3.client.autoscaling.update_auto_scaling_group(
                ctx,
                **{
                    "AutoScalingGroupName": name,
                    "LaunchConfigurationName": launch_configuration_name,
                    "LaunchTemplate": launch_template,
                    "MixedInstancesPolicy": mixed_instances_policy,
                    "InstanceId": instance_id,
                    "MinSize": min_size,
                    "MaxSize": max_size,
                    "DesiredCapacity": desired_capacity,
                    "DefaultCooldown": default_cooldown,
                    "AvailabilityZones": availability_zones,
                    "HealthCheckType": health_check_type,
                    "HealthCheckGracePeriod": health_check_grace_period,
                    "PlacementGroup": placement_group,
                    "VPCZoneIdentifier": vpc_zone_identifier,
                    "TerminationPolicies": termination_policies,
                    "NewInstancesProtectedFromScaleIn": new_instances_protected_from_scale_in,
                    "CapacityRebalance": capacity_rebalance,
                    "LifecycleHookSpecificationList": lifecycle_hook_specification_list,
                    "ServiceLinkedRoleARN": service_linked_role_arn,
                    "MaxInstanceLifetime": max_instance_lifetime,
                    "DesiredCapacityType": desired_capacity_type,
                },
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.autoscaling.auto_scaling_group", name=name
            )

        # Update tags
        if is_tags_updated:
            update_tags = await hub.exec.aws.autoscaling.tag.update_tags(
                ctx,
                old_tags=current_state["tags"],
                new_tags=desired_state["tags"],
                resource_id=resource_id,
            )
            result["comment"] = result["comment"] + update_tags["comment"]
            result["result"] = result["result"] and update_tags["result"]
            if not result["result"]:
                result["comment"] = update_tags["comment"]
                return result
    else:
        # Handle actual resource creation
        tags_list = hub.tool.aws.tag_utils.convert_auto_scaling_tag_dict_to_list(tags)
        translated_tags = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_present_tags_to_raw_tags(
            resource_id=name, tags=tags_list
        )
        ret = await hub.exec.boto3.client.autoscaling.create_auto_scaling_group(
            ctx,
            **{
                "AutoScalingGroupName": name,
                "LaunchConfigurationName": launch_configuration_name,
                "LaunchTemplate": launch_template,
                "MixedInstancesPolicy": mixed_instances_policy,
                "InstanceId": instance_id,
                "MinSize": min_size,
                "MaxSize": max_size,
                "DesiredCapacity": desired_capacity,
                "DefaultCooldown": default_cooldown,
                "AvailabilityZones": availability_zones,
                "LoadBalancerNames": load_balancer_names,
                "TargetGroupARNs": target_group_ar_ns,
                "HealthCheckType": health_check_type,
                "HealthCheckGracePeriod": health_check_grace_period,
                "PlacementGroup": placement_group,
                "VPCZoneIdentifier": vpc_zone_identifier,
                "TerminationPolicies": termination_policies,
                "NewInstancesProtectedFromScaleIn": new_instances_protected_from_scale_in,
                "CapacityRebalance": capacity_rebalance,
                "LifecycleHookSpecificationList": lifecycle_hook_specification_list,
                "Tags": translated_tags,
                "ServiceLinkedRoleARN": service_linked_role_arn,
                "MaxInstanceLifetime": max_instance_lifetime,
                "DesiredCapacityType": desired_capacity_type,
            },
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )
        resource_id = name

    # Populate new state
    new_auto_scaling_groups = (
        await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(
            ctx, AutoScalingGroupNames=[resource_id]
        )
    )
    if not new_auto_scaling_groups["result"]:
        result["comment"] = result["comment"] + new_auto_scaling_groups["comment"]
        result["result"] = new_auto_scaling_groups["result"]
        return result

    after = new_auto_scaling_groups["ret"]["AutoScalingGroups"][0]
    result[
        "new_state"
    ] = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
        auto_scaling_group=after
    )
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified Auto Scaling group. If the group has instances or scaling activities in progress, you must
    specify the option to force the deletion in order for it to succeed. If the group has policies, deleting the
    group deletes the policies, the underlying alarm actions, and any alarm that no longer has an associated action.
    To remove instances from the Auto Scaling group before deleting it, call the DetachInstances API with the list
    of instances and the option to decrement the desired capacity. This ensures that Amazon EC2 Auto Scaling does
    not launch replacement instances. To terminate all instances before deleting the Auto Scaling group, call the
    UpdateAutoScalingGroup API and set the minimum size and desired capacity of the Auto Scaling group to zero.

    Args:
        resource_id(Text, optional): The name of the Auto Scaling group.
        name(Text): An Idem name of the resource.
        timeout(Dict, optional): Timeout configuration for deletion of AWS auto scaling group.
            * delete (Dict) -- Timeout configuration for deletion of an auto scaling group
                * delay -- The amount of time in seconds to wait between attempts. Defaults to 15
                * max_attempts -- Customized timeout configuration containing delay and max attempts. Defaults to 40

    Request Syntax:
       [auto-scaling-group-resource-id]:
            aws.autoscaling.auto_scaling_group.absent:
              - resource_id: 'string'
              - name: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem-scaling-group:
              aws.autoscaling.auto_scaling_group.absent:
              - resource_id: idem-scaling-group
              - name: idem-scaling-group
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    # Query for auto scaling groups
    auto_scaling_group = None
    if resource_id:
        auto_scaling_group = (
            await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(
                ctx, AutoScalingGroupNames=[resource_id]
            )
        )

    # Check for existing auto scaling group
    before = None
    if (
        auto_scaling_group
        and auto_scaling_group["result"]
        and auto_scaling_group["ret"]["AutoScalingGroups"]
    ):
        before = auto_scaling_group["ret"]["AutoScalingGroups"][0]
        result[
            "old_state"
        ] = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
            auto_scaling_group=before
        )

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )

    elif ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )

    else:
        ret = await hub.exec.boto3.client.autoscaling.delete_auto_scaling_group(
            ctx,
            AutoScalingGroupName=resource_id,
            ForceDelete=True,
        )
        if not ret["result"]:
            result["result"] = ret["result"]
            result["comment"] = ret["comment"]
            return result

        # Custom waiter for delete
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        delete_waiter_acceptors = [
            {
                "matcher": "path",
                "expected": True,
                "state": "success",
                "argument": "length(AutoScalingGroups) == `0`",
            },
            {
                "matcher": "path",
                "expected": True,
                "state": "retry",
                "argument": "length(AutoScalingGroups) > `0`",
            },
        ]
        auto_scaling_group_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="AutoScalingGroupDelete",
            operation="DescribeAutoScalingGroups",
            argument=["AutoScalingGroups"],
            acceptors=delete_waiter_acceptors,
            client=hub.tool.boto3.client.get_client(ctx, "autoscaling"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "autoscaling",
                "AutoScalingGroupDelete",
                auto_scaling_group_waiter,
                AutoScalingGroupNames=[
                    resource_id,
                ],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.autoscaling.auto_scaling_group", name=name
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Gets information about the Auto Scaling groups in the account and Region.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.autoscaling.auto_scaling_group
    """

    result = {}
    ret = await hub.exec.boto3.client.autoscaling.describe_auto_scaling_groups(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe auto_scaling_group {ret['comment']}")
        return {}

    for auto_scaling_group in ret["ret"]["AutoScalingGroups"]:
        resource_id = auto_scaling_group["AutoScalingGroupName"]
        resource = hub.tool.aws.autoscaling.auto_scaling_group_utils.convert_raw_auto_scaling_group_to_present(
            auto_scaling_group=auto_scaling_group
        )
        result[resource_id] = {
            "aws.autoscaling.auto_scaling_group.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



from typing import Any
from typing import Dict
from typing import List


async def update_tags(
    hub,
    ctx,
    resource_id: str,
    old_tags: List[Dict[str, Any]],
    new_tags: List[Dict[str, Any]],
):
    """

    Args:
        resource_id: aws organizations resource id
        old_tags: List of existing tags
        new_tags: List of new tags

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}

    """

    result = dict(comment=(), result=True, ret=None)

    old_tags_map = {tag.get("Key"): tag.get("Value") for tag in old_tags}
    new_tags_map = {tag.get("Key"): tag.get("Value") for tag in new_tags}

    if old_tags_map == new_tags_map:
        result["comment"] = result["comment"] + ("Tag update is not required",)
        return result

    tags_to_add = []
    tags_to_delete = []

    for key, value in new_tags_map.items():
        if (key in old_tags_map and old_tags_map.get(key) != new_tags_map.get(key)) or (
            key not in old_tags_map
        ):
            tags_to_add.append({"Key": key, "Value": value})

    for key in old_tags_map:
        if key not in new_tags_map:
            tags_to_delete.append(key)
    if not ctx.get("test", False):
        try:
            delete_tag_resp = await hub.exec.boto3.client.organizations.untag_resource(
                ctx, ResourceId=resource_id, TagKeys=tags_to_delete
            )

            if not delete_tag_resp["result"]:
                hub.log.debug(
                    f"Could not delete tags {tags_to_delete} on resource {resource_id} with error {delete_tag_resp['comment']}"
                )
                result["comment"] = result["comment"] + delete_tag_resp["comment"]
                result["result"] = delete_tag_resp["result"]
                return result

        except hub.tool.boto3.exception.ClientError as e:
            hub.log.debug(
                f"Could not delete tags {tags_to_delete} on resource {resource_id} with error {e}"
            )
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

        try:
            create_tag_resp = await hub.exec.boto3.client.organizations.tag_resource(
                ctx, ResourceId=resource_id, Tags=tags_to_add
            )
            if not create_tag_resp["result"]:
                hub.log.debug(
                    f"Could not create tags {tags_to_add} on resource {resource_id} with error {create_tag_resp['comment']}"
                )
                result["comment"] = result["comment"] + create_tag_resp["comment"]
                result["result"] = create_tag_resp["result"]
                return result

        except hub.tool.boto3.exception.ClientError as e:
            hub.log.debug(
                f"Error while creating tags {tags_to_add} on resource {resource_id} : {e}"
            )
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    result["comment"] = result["comment"] + (f"Updated tags '{new_tags}'",)
    result["ret"] = new_tags
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS Elasticache to present input format.
"""


async def convert_raw_elasticache_subnet_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """
    Return converted raw resource state from AWS Elasticache subnet group to required input format for present.

     Args:
        raw_resource(Text): resource obtained from describe API
        idem_resource_name(Text): name of idem resource

    Return: Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)
    resource_id = raw_resource.get("CacheSubnetGroupName")
    resource_parameters = OrderedDict(
        {
            "CacheSubnetGroupDescription": "cache_subnet_group_description",
            "ARN": "arn",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("Subnets"):
        subnet_ids_list = []
        for subnet in raw_resource.get("Subnets"):
            if "SubnetIdentifier" in subnet:
                subnet_ids_list.append(subnet.get("SubnetIdentifier"))
        resource_translated["subnet_ids"] = subnet_ids_list

    if raw_resource.get("ARN"):
        tags = await hub.exec.boto3.client.elasticache.list_tags_for_resource(
            ctx, ResourceName=raw_resource.get("ARN")
        )
        result["result"] = tags["result"]
        if tags["result"] and tags.get("ret"):
            resource_translated[
                "tags"
            ] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags["ret"]["TagList"])
        if not result["result"]:
            result["comment"] = result["comment"] + tags["comment"]
    result["ret"] = resource_translated
    return result


async def convert_raw_elasticache_parameter_group_to_present(
    hub,
    ctx,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
) -> Dict[str, Any]:
    """
    Return converted raw resource state from AWS Elasticache parameter group to required input format for present.

     Args:
        raw_resource(Text): resource obtained from describe API
        idem_resource_name(Text): name of idem resource

    Return: Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)
    resource_id = raw_resource.get("CacheParameterGroupName")
    resource_parameters = OrderedDict(
        {
            "CacheParameterGroupFamily": "cache_parameter_group_family",
            "Description": "description",
            "ARN": "arn",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    if raw_resource.get("ARN"):
        tags = await hub.exec.boto3.client.elasticache.list_tags_for_resource(
            ctx, ResourceName=raw_resource.get("ARN")
        )
        result["result"] = tags["result"]
        if tags["result"] and tags.get("ret"):
            resource_translated[
                "tags"
            ] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags["ret"]["TagList"])
        if not tags["result"]:
            result["comment"] = result["comment"] + tags["comment"]

    ret_parameters = await hub.exec.boto3.client.elasticache.describe_cache_parameters(
        ctx, CacheParameterGroupName=resource_id
    )
    result["result"] = result["result"] and ret_parameters["result"]
    if not ret_parameters["result"]:
        result["comment"] = result["comment"] + ret_parameters["comment"]
    elif ret_parameters["result"] and ret_parameters.get("ret"):
        updated_parameter_list = []
        for parameter in ret_parameters["ret"].get("Parameters"):
            parameter_list = {}
            if "ParameterName" in parameter:
                parameter_list["ParameterName"] = parameter.get("ParameterName")
            if "ParameterValue" in parameter:
                parameter_list["ParameterValue"] = parameter.get("ParameterValue")
            updated_parameter_list.append(parameter_list)
        resource_translated["parameter_name_values"] = updated_parameter_list

    result["ret"] = resource_translated
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List


async def update_anomaly_subscription(
    hub,
    ctx,
    before: Dict[str, Any],
    subscription_arn: str,
    threshold: float,
    frequency: str,
    monitor_arn_list: List,
    subscribers: List,
    subscription_name: str,
):
    """
    Updates the Anomaly Subscription
    Args:
        before: existing resource
        subscription_arn (Text) -- ARN of the subscription
        threshold (float) -- The dollar value that triggers a notification if the threshold is exceeded.
        frequency (Text) -- The frequency that anomaly reports are sent over email.
        monitor_arn_list (List) -- An Idem list of cost anomaly monitors.
        subscribers (List) -- A list of subscribers to notify.
        subscription_name: Name of the subscription

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": None}

    """
    result = dict(comment=(), result=True, ret=None)
    update_payload = {}

    resource_parameters = OrderedDict(
        {
            "Threshold": threshold,
            "Frequency": frequency,
            "SubscriptionName": subscription_name,
        }
    )

    if not hub.tool.aws.state_comparison_utils.are_lists_identical(
        monitor_arn_list,
        before.get("MonitorArnList"),
    ):
        update_payload["MonitorArnList"] = monitor_arn_list

    if not hub.tool.aws.state_comparison_utils.are_lists_identical(
        subscribers,
        before.get("Subscribers"),
    ):
        update_payload["Subscribers"] = subscribers

    for key, value in resource_parameters.items():
        if key in before.keys():
            if (value is not None) and value != before[key]:
                update_payload[key] = resource_parameters[key]

    if update_payload:
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.ce.update_anomaly_subscription(
                ctx, SubscriptionArn=subscription_arn, **update_payload
            )
            if not update_ret["result"]:
                result["comment"] = result["comment"] + update_ret["comment"]
                result["result"] = False
                return result
        result["ret"] = {}
        result = update_result(result, update_payload)

    return result


def update_result(
    result: Dict[str, Any], update_payload: Dict[str, Any]
) -> Dict[str, Any]:
    parameters = OrderedDict(
        {
            "MonitorArnList": "monitor_arn_list",
            "Subscribers": "subscribers",
            "Threshold": "threshold",
            "Frequency": "frequency",
            "SubscriptionName": "subscription_name",
        }
    )

    for raw_parameter, present_parameter in parameters.items():
        if raw_parameter in update_payload:
            result["ret"][present_parameter] = update_payload[raw_parameter]
            result["comment"] = result["comment"] + (
                f"Update {present_parameter}: {update_payload[raw_parameter]}",
            )
    return result



from collections import OrderedDict
from typing import Any
from typing import Dict


def convert_raw_resource_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    rest_api_id: str = None,
) -> Dict[str, Any]:
    r"""
    Convert AWS API Gateway Resource to a common idem present state.

    Args:
        hub: required for functions in hub.
        raw_resource(Dict[str, Any]): The AWS response to convert.
        idem_resource_name(str): Optional parameter -- will be the resource name if applicable,
            otherwise resource name will be taken from raw_resource.
        rest_api_id(str): Rest Api Id associated with the Rest Api to which the Resource belongs.

    Returns:
        Dict[str, Any]: Common idem present state
    """

    resource_id = raw_resource.get("id")

    if idem_resource_name:
        name = idem_resource_name
    else:
        name = raw_resource.get("pathPart")

    resource_parameters = OrderedDict(
        {
            "parentId": "parent_id",
            "pathPart": "path_part",
        }
    )

    resource_translated = {
        "name": name,
        "resource_id": resource_id,
        "rest_api_id": rest_api_id,
    }
    for parameter_raw, parameter_present in resource_parameters.items():
        resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    return resource_translated



"""
Plugin for aws enforced state management
"""
import json
import pathlib
from typing import Any
from typing import Dict

import botocore.exceptions


def __init__(hub):
    hub.esm.aws.ACCT = ["aws"]
    hub.esm.aws.ESM_PROFILE = {}


async def get_state(hub, ctx) -> Dict[str, Any]:
    """
    Read this context's state from AWS S3 bucket. Return empty dict if the file key path does not exist in the bucket.

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.

    Returns:
        A dict containing all resource state records. Return empty dict if the file key path does not exist in the bucket.
    """
    try:
        client = hub.tool.boto3.client.get_client(ctx, service_name="s3")
        content = client.get_object(
            Bucket=hub.esm.aws.ESM_PROFILE.get("bucket"),
            Key=hub.esm.aws.ESM_PROFILE.get("key"),
        )
        state = json.loads(content["Body"].read())
        return state
    except botocore.exceptions.ClientError as client_ex:
        if client_ex.response["Error"]["Code"] == "NoSuchKey":
            hub.log.debug(
                f"No state file found from S3 bucket {hub.esm.aws.ESM_PROFILE.get('bucket')} using key {hub.esm.aws.ESM_PROFILE.get('key')}"
            )
            return {}
        else:
            hub.log.error(
                f"Failed to retrieve state file from S3 bucket {hub.esm.aws.ESM_PROFILE.get('bucket')} using key {hub.esm.aws.ESM_PROFILE.get('key')}"
                f" with error {client_ex.response['Error']['Message']}"
            )
            raise client_ex
    except Exception as e:
        hub.log.error(
            f"Failed to retrieve state file from S3 bucket {hub.esm.aws.ESM_PROFILE.get('bucket')} using key {hub.esm.aws.ESM_PROFILE.get('key')}"
            f" with error {str(e)}"
        )
        raise e


async def set_state(hub, ctx, state: Dict[str, Any]):
    """
    Write the state to this context's cache_file

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        state: A dict containing all resource states that will be stored in S3 bucket.

    Returns:
        None if the operation succeeds. An exception will raise if the operation failed.
    """
    try:
        ret = await hub.exec.boto3.client.s3.put_object(
            ctx,
            Bucket=hub.esm.aws.ESM_PROFILE.get("bucket"),
            Body=bytes(json.dumps(state), "utf-8"),
            Key=hub.esm.aws.ESM_PROFILE.get("key"),
            ServerSideEncryption="AES256"
            if hub.esm.aws.ESM_PROFILE.get("encrypt", False)
            else None,
        )
        if not ret["result"]:
            error_msg = (
                f"Failed to upload state file to S3 bucket {hub.esm.aws.ESM_PROFILE.get('bucket')} as"
                f" {hub.esm.aws.ESM_PROFILE.get('key')} with error {ret['comment']}"
            )
            raise RuntimeError(error_msg)
        else:
            hub.log.debug(
                f"Successfully upload state file to S3 bucket {hub.esm.aws.ESM_PROFILE.get('bucket')} as"
                f" {hub.esm.aws.ESM_PROFILE.get('key')}"
            )
            return
    except Exception as e:
        hub.log.error(str(e))
        raise e


async def enter(hub, ctx):
    """
    Verify that only one process can access the same enforced state record file in S3 bucket.
    Set a lock by performing an atomic operation to put the file key to DynamoDB.
    If esm profile is not specified in credential profile. Then we try to find it from the Idem config.
    The config file should have the following structure:

      extras:
        aws:
          esm:
            bucket: idem-storage-bucket-name
            dynamodb_table: idem-esm-lock-table-name
            key: /my-file-path/my-esm-cache.json
            encrypt: True

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.

    Returns:
        None if lock has been acquired successfully. Raise exception if lock was failed to be acquired
    """
    hub.esm.aws.ESM_PROFILE = ctx.acct.get("esm")
    if not hub.esm.aws.ESM_PROFILE:
        if (
            "extras" in hub.OPT.acct
            and hub.OPT.acct.extras
            and "aws" in hub.OPT.acct.extras
        ):
            hub.esm.aws.ESM_PROFILE = hub.OPT.acct.extras.aws.get("esm")
    if not hub.esm.aws.ESM_PROFILE:
        raise ValueError("ESM profile is required to support remote ESM storage.")
    if not hub.esm.aws.ESM_PROFILE.get("dynamodb_table"):
        raise ValueError("dynamodb_table is required in ESM profile.")
    if not hub.esm.aws.ESM_PROFILE.get("bucket"):
        raise ValueError("bucket is required in ESM profile.")
    if not hub.esm.aws.ESM_PROFILE.get("key"):
        raise ValueError("key is required in ESM profile.")
    try:
        # Put item with conditional expression to acquire the lock. The conditional check and put operation is atomic.
        # This ensures that the S3 bucket file can be locked properly.
        db_resource = hub.tool.boto3.resource.create(
            ctx, "dynamodb", "Table", hub.esm.aws.ESM_PROFILE.get("dynamodb_table")
        )
        db_resource.put_item(
            Item={
                "LockID": f"{hub.esm.aws.ESM_PROFILE.get('bucket')}/{hub.esm.aws.ESM_PROFILE.get('key')}"
            },
            ConditionExpression="attribute_not_exists(#r)",
            ExpressionAttributeNames={"#r": "LockID"},
        )
        # Lock acquired
        return
    except botocore.exceptions.ClientError as e:
        # Another exception than ConditionalCheckFailedException was caught, raise as-is
        if e.response["Error"]["Code"] != "ConditionalCheckFailedException":
            raise e
        else:
            # If conditional check failed, it means the lock cannot be acquired because it is already locked
            raise ValueError(
                f"File {hub.esm.aws.ESM_PROFILE.get('bucket')}/{hub.esm.aws.ESM_PROFILE.get('key')}"
                f" has been locked by another user/process."
            )


async def exit_(hub, ctx, handle: pathlib.Path, exception: Exception):
    """
    Release the db lock and handle errors

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        handle: This parameter is inherited from Idem. It is not being used in Idem-aws plugin.
        exception: Error.

    Returns:
        None if file lock has been released. Raise Exception if any failure happened.
    """
    if exception:
        hub.log.error(f"{exception.__class__.__name__}: {exception}")
    if hub.esm.aws.ESM_PROFILE is None:
        raise ValueError("ESM profile is required to support remote ESM storage.")
    try:
        # Delete item with conditional expression to release the lock
        db_resource = hub.tool.boto3.resource.create(
            ctx, "dynamodb", "Table", hub.esm.aws.ESM_PROFILE.get("dynamodb_table")
        )
        db_resource.delete_item(
            Key={
                "LockID": f"{hub.esm.aws.ESM_PROFILE.get('bucket')}/{hub.esm.aws.ESM_PROFILE.get('key')}"
            },
            ConditionExpression="attribute_exists(#r)",
            ExpressionAttributeNames={"#r": "LockID"},
        )
        # Lock released
        return
    except botocore.exceptions.ClientError as e:
        # Another exception than ConditionalCheckFailedException was caught, raise as-is
        if e.response["Error"]["Code"] != "ConditionalCheckFailedException":
            raise e
        else:
            # Else, lock has been released
            hub.log.debug(
                f"Lock {hub.esm.aws.ESM_PROFILE.get('bucket')}/{hub.esm.aws.ESM_PROFILE.get('key')}"
                f" has been previously released."
            )
            return



def get_updated_cache_parameter_group(hub, old_paramaters, parameter_name_values):

    r"""
    Checks if the input ParameterNamesValues needs to be updated and returns a list of parameters that needs to be modified.

    Args:

        old_parameters: The detailed parameter list for a particular cache parameter group.
        parameter_name_values: Parameters Values as needed by the user.

    Returns:
        [
            {
                "ParameterName": "Text",
                "ParameterValue": "Text"
            },
        ]
    """
    final_parameters = []
    old_parameter_map = {
        params.get("ParameterName"): params for params in old_paramaters
    }
    if parameter_name_values:
        for new_parameters in parameter_name_values:
            if new_parameters.get("ParameterName") in old_parameter_map:
                if not new_parameters.get("ParameterValue") == old_parameter_map.get(
                    new_parameters.get("ParameterName")
                ).get("ParameterValue"):
                    final_parameters.append(new_parameters)
    return final_parameters



import asyncio
import re
from typing import Any
from typing import Dict
from typing import Generator

import boto3.session

ITERATION_FINISHED = object()

__func_alias__ = {"exec_": "exec"}


async def exec_(
    hub,
    ctx,
    service_name: str,
    operation: str,
    jmes_search_path: str = None,
    *op_args,
    **op_kwargs: Dict[str, Any],
) -> Any:
    """
    :param hub:
    :param ctx:
    :param service_name: The name of the service client to create
    :param operation: The operation to run from the service client
    :param jmes_search_path: The JMES path to use as a filter for paginated results
    :param op_args: arguments to pass to the operation call
    :param op_kwargs: keyword arguments to pass to the operation call

    :return: The result of the operation call
    """
    session: boto3.session.Session = hub.tool.boto3.session.get()
    client = session.client(
        service_name=service_name,
        region_name=ctx.acct.get("region_name"),
        api_version=ctx.acct.get("api_version"),
        use_ssl=ctx.acct.get("use_ssl", True),
        endpoint_url=ctx.acct.get("endpoint_url"),
        aws_access_key_id=ctx.acct.get("aws_access_key_id"),
        aws_secret_access_key=ctx.acct.get("aws_secret_access_key"),
        aws_session_token=ctx.acct.get("aws_session_token"),
        verify=ctx.acct.get("verify"),
    )

    # Don't pass kwargs that have a "None" value to the function call
    kwargs = {k: v for k, v in op_kwargs.items() if v is not None}

    can_paginate = client.can_paginate(operation)

    if can_paginate:
        hub.log.debug(f"Paginating results for {service_name}.{operation}")
        paginator = client.get_paginator(operation)
        pages = paginator.paginate(*op_args, **kwargs)
        if jmes_search_path is None:
            return await hub.pop.loop.wrap(pages.build_full_result)
        else:
            iterator = await hub.pop.loop.wrap(pages.search, jmes_search_path)
            return [_ for _ in iterator]
    else:
        hub.log.debug(f"Getting raw results for {service_name}.{operation}")
        op = getattr(client, operation)
        return await hub.pop.loop.wrap(op, *op_args, **kwargs)


def get_client(hub, ctx, service_name: str):
    session: boto3.session.Session = hub.tool.boto3.session.get()
    client = session.client(
        service_name=service_name,
        region_name=ctx.acct.get("region_name"),
        api_version=ctx.acct.get("api_version"),
        use_ssl=ctx.acct.get("use_ssl", True),
        endpoint_url=ctx.acct.get("endpoint_url"),
        aws_access_key_id=ctx.acct.get("aws_access_key_id"),
        aws_secret_access_key=ctx.acct.get("aws_secret_access_key"),
        aws_session_token=ctx.acct.get("aws_session_token"),
        verify=ctx.acct.get("verify"),
    )
    return client


async def wait(
    hub,
    ctx,
    service_name: str,
    waiter_name: str,
    custom_waiter=None,
    delay: int = None,
    *wt_args,
    **wt_kwargs,
) -> None:
    """
    Asynchronously wait for the named resource to be available

    :param hub:
    :param ctx:
    :param service_name: The name of the service client to retrieve
    :param waiter_name: The name of the waiter to get from the service client
    :param custom_waiter: If an inbuilt waiter is not available , we can provide custom waiter for the service client
    :param delay: The delay in seconds before starting the waiter
    :param wt_args: Args to pass to the wait function
    :param wt_kwargs: kwargs to pass to the wait function
    """

    if custom_waiter is not None:
        waiter = custom_waiter

        if waiter is None:
            raise NameError(f"No custom waiter object defined for '{waiter_name}'. ")
    else:
        client = hub.tool.boto3.client.get_client(ctx, service_name)

        if waiter_name not in client.waiter_names:
            raise NameError(
                f"No waiter '{waiter_name}'. "
                f"Available waiters for '{service_name}' are: {' '.join(client.waiter_names)}"
            )

        waiter = client.get_waiter(waiter_name)

    if delay is not None:
        # There might be some delay for the resource to start updating/deleting after the api call
        hub.log.debug(f"Waiting '{delay}' seconds, before starting the waiter")
        await asyncio.sleep(delay)

    await hub.pop.loop.wrap(waiter.wait, *wt_args, **wt_kwargs)


def search(hub, ctx, service_name: str, collection: str, jmes_path: str = "*[]") -> str:
    """
    :param hub:
    :param ctx:
    :param service_name:
    :param collection:
    :param jmes_path:
    :return:
    """
    session: boto3.session.Session = hub.tool.boto3.session.get()
    client = session.client(
        service_name=service_name,
        region_name=ctx.acct.get("region_name"),
        api_version=ctx.acct.get("api_version"),
        use_ssl=ctx.acct.get("use_ssl", True),
        endpoint_url=ctx.acct.get("endpoint_url"),
        aws_access_key_id=ctx.acct.get("aws_access_key_id"),
        aws_secret_access_key=ctx.acct.get("aws_secret_access_key"),
        aws_session_token=ctx.acct.get("aws_session_token"),
        verify=ctx.acct.get("verify"),
    )

    # Find a paginator that can describe this collection
    for operation in client.meta.method_to_api_mapping:
        if client.can_paginate(operation) and re.match(
            f"[a-z]+_{collection.lower()}s?$", operation
        ):
            break
    else:
        raise AttributeError(
            f"Could not find a paginator for {service_name}.{collection}"
        )

    paginator = client.get_paginator(operation)
    pages = paginator.paginate()
    iterator: Generator = pages.search(jmes_path)
    try:
        return next(iterator, None)
    finally:
        iterator.close()



import json


def get_default_topic_policy(hub, topic_arn: str) -> str:
    default_policy = {
        "Version": "2008-10-17",
        "Id": "__default_policy_ID",
        "Statement": [
            {
                "Sid": "__default_statement_ID",
                "Effect": "Allow",
                "Principal": {"AWS": "*"},
                "Action": [
                    "SNS:GetTopicAttributes",
                    "SNS:SetTopicAttributes",
                    "SNS:AddPermission",
                    "SNS:RemovePermission",
                    "SNS:DeleteTopic",
                    "SNS:Subscribe",
                    "SNS:ListSubscriptionsByTopic",
                    "SNS:Publish",
                ],
                "Resource": "topic_arn",
                "Condition": {"StringEquals": {"AWS:SourceOwner": "acct_id"}},
            }
        ],
    }
    default_policy["Statement"][0]["Resource"] = topic_arn
    acct_id = topic_arn.split(":")[4]
    default_policy["Statement"][0]["Condition"]["StringEquals"][
        "AWS:SourceOwner"
    ] = acct_id
    policy = json.dumps(default_policy, separators=(", ", ": "))
    return policy



import copy
from typing import Any
from typing import Dict

"""
Exec functions for AWS API Gateway tags.
"""


async def update_tags(
    hub,
    ctx,
    resource_arn: str,
    old_tags: Dict[str, Any],
    new_tags: Dict[str, Any],
):
    r"""
    Updates tags for an AWS API Gateway resource.

    Args:
        hub: required for functions in hub.
        ctx: context.
        resource_arn: AWS API Gateway resource arn.
        old_tags(Dict[str, Any]): Dict of existing tags.
        new_tags(Dict[str, Any]): Dict of new tags.

    Returns:
        {"result": True|False, "comment": A message Tuple, "ret": dict of updated tags}
    """

    result = dict(comment=(), result=True, ret=None)

    if new_tags is not None:
        tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
            old_tags=old_tags, new_tags=new_tags
        )
    if (not tags_to_remove) and (not tags_to_add):
        result["ret"] = copy.deepcopy(old_tags if old_tags else {})
        return result

    if tags_to_remove:
        if not ctx.get("test", False):
            untag_ret = await hub.exec.boto3.client.apigateway.untag_resource(
                ctx,
                resourceArn=resource_arn,
                tagKeys=list(tags_to_remove.keys()),
            )
            if not untag_ret["result"]:
                result["result"] = False
                result["comment"] += untag_ret["comment"]
                return result

    if tags_to_add:
        if not ctx.get("test", False):
            tag_ret = await hub.exec.boto3.client.apigateway.tag_resource(
                ctx,
                resourceArn=resource_arn,
                tags=tags_to_add,
            )
            if not tag_ret["result"]:
                result["result"] = False
                result["comment"] += tag_ret["comment"]
                return result

    result["ret"] = new_tags
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    else:
        result["comment"] = hub.tool.aws.comment_utils.update_tags_comment(
            tags_to_remove=tags_to_remove, tags_to_add=tags_to_add
        )
    return result



"""
hub.exec.boto3.client.config.put_config_rule
hub.exec.boto3.client.config.delete_config_rule
hub.exec.boto3.client.config.describe_config_rules

"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    scope: Dict = None,
    source: Dict = True,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    max_execution_frequency: str = None,
    input_parameters: str = None,
) -> Dict[str, Any]:
    """
    Adds or updates Config rule for evaluating whether your Amazon Web Services resources comply with your desired
    configurations. For more information about rules, please see AWS Config services
    Args:
        name(Text): An Idem name of the rule.
        resource_id(Text, Optional): AWS Config Rule Name.
        scope(Dict, Optional): Defines which resources can trigger an evaluation for the rule. The scope can include one or more
         resource types, a combination of one resource type and one resource ID, or a combination of a tag key and value
        source(Dict): Provides the rule owner, the rule identifier, and the
         notifications that cause the function to evaluate your Amazon Web Services resources,
        max_execution_frequency(Text, Optional): The maximum frequency with which Config runs evaluations for a rule. Default
         is 24 hours
        input_parameters(Text, Optional): A string, in JSON format, that is passed to the Config rule.
        tags (Dict or List, Optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the Config rule.
            * (Key, optional): One part of a key-value pair that make up a tag. A key is a general label that acts like a
                category for more specific tag values.
            * (Value, optional): The optional part of a key-value pair that make up a tag. A value acts as a descriptor within a
                tag category (key).

        tags(List or Dict, Optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
         {tag-key: tag-value}. The tags for the resource. The metadata that you apply to a resource to help you categorize and
         organize them. Defaults to None

    Request syntax:
        [aws-config-rule]:
          aws.config.rule.present:
          - name: 'string'
          - resource_id: 'string'
          - scope: dict
            ComplianceResourceTypes: list
          - source: dict
            Owner: 'string'
            SourceIdentifier: 'string'

    Returns:
         None

    Examples:
        .. code-block:: sls

            ec2-instance-no-public-ip:
              aws.config.rule.present:
              - name: ec2-instance-no-public-ip
              - resource_id: ec2-instance-no-public-ip
              - tags:
                - Key: ENV
                  Value: Test
                - Key: Service
                  Value: TestService
              - config_rule_name: ec2-instance-no-public-ip
              - scope:
                  ComplianceResourceTypes:
                  - AWS::EC2::Instance
                  - AWS::EC2::Host
              - source:
                  Owner: AWS
                  SourceIdentifier: EC2_INSTANCE_NO_PUBLIC_IP
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    update_scope = None
    if resource_id:
        try:
            before = await hub.exec.boto3.client.config.describe_config_rules(
                ctx, ConfigRuleNames=[resource_id]
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
            return result
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
                ctx,
                raw_resource=before["ret"]["ConfigRules"][0],
                idem_resource_name=name,
            )
            plan_state = copy.deepcopy(result["old_state"])
            # updating scope
            if scope is not None:
                update_scope = (
                    scope if scope is not None else result["old_state"]["scope"]
                )

            update_ret = await hub.exec.aws.config.rule.update_rule(
                ctx,
                resource_id=resource_id,
                before=before["ret"]["ConfigRules"][0],
                source=source,
                scope=update_scope,
                frequency=max_execution_frequency,
                input_parameters=input_parameters,
            )

            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                for key in ["max_execution_frequency", "scope", "input_parameters"]:
                    if key in update_ret["ret"]:
                        plan_state[key] = update_ret["ret"].get(key)

            if (tags is not None) and tags != result["old_state"].get("tags"):
                update_tag_ret = await hub.exec.aws.config.tag.update_tags(
                    ctx,
                    result["old_state"]["config_rule_arn"],
                    result["old_state"]["tags"],
                    tags,
                )
                result["result"] = result["result"] and update_tag_ret["result"]
                result["comment"] = result["comment"] + update_tag_ret["comment"]
                resource_updated = resource_updated or bool(update_tag_ret["result"])
                if ctx.get("test", False) and update_tag_ret["ret"] is not None:
                    plan_state["tags"] = update_tag_ret["ret"]
            if not resource_updated:
                result["comment"] = result["comment"] + (f"{name} already exists",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    else:
        try:
            if ctx.get("test", False):
                desired_state_payload = {
                    "name": name,
                    "config_rule_name": name,
                    "source": source,
                    "tags": tags,
                }
                if scope:
                    desired_state_payload["scope"] = scope
                if input_parameters:
                    desired_state_payload["input_parameters"] = input_parameters
                if max_execution_frequency:
                    desired_state_payload[
                        "max_execution_frequency"
                    ] = max_execution_frequency
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={}, desired_state=desired_state_payload
                )
                result["comment"] = (f"Would create aws.config.rule {name}",)
                return result
            config_rule_payload = {"ConfigRuleName": name, "Source": source}
            if scope:
                config_rule_payload["Scope"] = scope
            if max_execution_frequency:
                config_rule_payload[
                    "MaximumExecutionFrequency"
                ] = max_execution_frequency
            if input_parameters:
                config_rule_payload["InputParameters"] = input_parameters
            ret = await hub.exec.boto3.client.config.put_config_rule(
                ctx,
                ConfigRule=config_rule_payload,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"] + ret["comment"]
                return result
            resource_id = name
            result["comment"] = result["comment"] + (f"Created '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}")
            result["result"] = False
    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.config.describe_config_rules(
                ctx, ConfigRuleNames=[resource_id]
            )
            result[
                "new_state"
            ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
                ctx,
                raw_resource=after["ret"]["ConfigRules"][0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Deletes the specified Config rule and all of its evaluation results. Config sets the state of a rule to DELETING
    until the deletion is complete. You cannot update a rule while it is in this state. If you make a PutConfigRule or
    DeleteConfigRule request for the rule, you will receive a ResourceInUseException.

    Args:
        name(Text): An Idem name of the rule.
        resource_id(Text, optional): AWS Config Rule Name. Idem automatically considers this resource being absent
         if this field is not specified.

    Returns:
          Dict[str, Any]

    Examples:
          .. code-block:: sls

            ec2-instance-no-public-ip:
              aws.config.rule.absent:
              - name: ec2-instance-no-public-ip
              - resource_id: ec2-instance-no-public-ip

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.rule", name=name
        )
        return result
    before = await hub.exec.boto3.client.config.describe_config_rules(
        ctx, ConfigRuleNames=[resource_id]
    )
    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.config.rule", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
            ctx, raw_resource=before["ret"]["ConfigRules"][0], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.config.rule", name=name
        )
        return result
    else:
        try:
            if before["ret"]["ConfigRules"][0]["ConfigRuleState"] == "DELETING":
                result["comment"] = (
                    f"aws.config.rule '{name}' is still in deleting state.",
                )
            else:
                result[
                    "old_state"
                ] = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
                    ctx,
                    raw_resource=before["ret"]["ConfigRules"][0],
                    idem_resource_name=name,
                )

                ret = await hub.exec.boto3.client.config.delete_config_rule(
                    ctx, ConfigRuleName=resource_id
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    result["result"] = False
                    return result
                result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                    resource_type="aws.config.rule", name=name
                )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Return details about your Config rules.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.config.rule
    """
    result = {}
    ret = await hub.exec.boto3.client.config.describe_config_rules(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe Rules {ret['comment']}")
        return {}

    for resource in ret["ret"]["ConfigRules"]:
        resource_id = resource.get("ConfigRuleName")
        resource_translated = await hub.tool.aws.config.conversion_utils.convert_raw_config_rule_to_present(
            ctx, raw_resource=resource, idem_resource_name=resource_id
        )
        result[resource_id] = {
            "aws.config.rule.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.eks.create_nodegroup
hub.exec.boto3.client.eks.delete_nodegroup
hub.exec.boto3.client.eks.describe_nodegroup
hub.exec.boto3.client.eks.list_nodegroups
"""
import copy
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]

TREQ = {
    "absent": {
        "require": [
            "aws.eks.cluster.absent",
        ],
    },
    "present": {
        "require": [
            "aws.eks.cluster.present",
        ],
    },
}
default_attempts = 120


async def present(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    subnets: List,
    node_role: str,
    resource_id: str = None,
    scaling_config: Dict = None,
    disk_size: int = None,
    instance_types: List = None,
    ami_type: str = None,
    remote_access: Dict = None,
    labels: Dict = None,
    taints: List = None,
    tags: Dict = None,
    client_request_token: str = None,
    launch_template: Dict = None,
    update_config: Dict = None,
    capacity_type: str = None,
    version: str = None,
    release_version: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a managed node group for an Amazon EKS cluster. You can only create a node group for your cluster that
    is equal to the current Kubernetes version for the cluster. All node groups are created with the latest AMI
    release version for the respective minor Kubernetes version of the cluster, unless you deploy a custom AMI using
    a launch template. For more information about using launch templates, see Launch template support. An Amazon EKS
    managed node group is an Amazon EC2 Auto Scaling group and associated Amazon EC2 instances that are managed by
    Amazon Web Services for an Amazon EKS cluster. Each node group uses a version of the Amazon EKS optimized Amazon
    Linux 2 AMI. For more information, see Managed Node Groups in the Amazon EKS User Guide.

    Args:
        name(Text): An Idem name of EKS Node group.
        cluster_name(Text): The name of the cluster to create the node group in.
        subnets(List): The subnets to use for the Auto Scaling group that is created for your node group.
        node_role(Text): The Amazon Resource Name (ARN) of the IAM role to associate with your node group.
            The Amazon EKS worker node kubelet daemon makes calls to Amazon Web Services APIs on your behalf. Nodes
            receive permissions for these API calls through an IAM instance profile and associated policies. Before you
            can launch nodes and register them into a cluster, you must create an IAM role for those nodes to use when
            they are launched.
        resource_id(Text, optional): AWS EKS Node group name
        scaling_config(Dict, optional): The scaling configuration details for the Auto Scaling group that is created
            for your node group
        capacity_type(Text, optional): The capacity type for your node group
        disk_size(integer, optional): The root device disk size (in GiB) for your node group instances. The default
            disk size is 20 GiB.
        instance_types(Text, optional): Specify the instance types for a node group.
        ami_type(Text, optional): The AMI type for your node group. GPU instance types should use the AL2_x86_64_GPU
            AMI type. Non-GPU instances should use the AL2_x86_64 AMI type. Arm instances should use the AL2_ARM_64 AMI
            type.
        remote_access(Dict, optional): The remote access (SSH) configuration to use with your node group. If you
            specify launchTemplate , then don't specify remoteAccess , or the node group deployment will fail.
        labels(Dict, optional): The Kubernetes labels to be applied to the nodes in the node group when they are created.
        taints(List, optional): The Kubernetes taints to be applied to the nodes in the node group.
        tags(Dict, optional): The metadata to apply to the node group to assist with categorization and organization.
            Each tag consists of a key and an optional value. You define both. Node group tags do not propagate to any
            other resources associated with the node group, such as the Amazon EC2 instances or subnets.
        client_request_token(Text, optional): Unique, case-sensitive identifier that you provide to ensure the
            idempotency of the request.
        launch_template(Dict, optional): An object representing a node group's launch template specification. If
            specified, then do not specify instanceTypes , diskSize , or remoteAccess and make sure that the launch
            template meets the requirements in launchTemplateSpecification
        update_config(Dict, optional): The node group update configuration
        version(Text, optional): The Kubernetes version to use for your managed nodes. By default, the Kubernetes
            version of the cluster is used, and this is the only accepted specified value. If you specify launchTemplate
            , and your launch template uses a custom AMI, then don't specify version , or the node group deployment will
            fail. For more information about using launch templates with Amazon EKS, see Launch template support in the
            Amazon EKS User Guide.
        release_version(Text, optional): The AMI version of the Amazon EKS optimized AMI to use with your node group.
            By default, the latest available AMI version for the node group's current Kubernetes version is used. For
            more information, see Amazon EKS optimized Amazon Linux 2 AMI versions in the Amazon EKS User Guide . If you
            specify launchTemplate , and your launch template uses a custom AMI, then don't specify releaseVersion ,
            or the node group deployment will fail. For more information about using launch templates with Amazon EKS,
            see Launch template support in the Amazon EKS User Guide.
        timeout(Dict, optional): Timeout configuration for creating or updating nodegroup.
            * create (Dict) -- Timeout configuration for creating nodegroup
                * delay(int, default=30) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=80) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating nodegroup
                * delay(int, default=30) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=80) -- Customized timeout configuration containing delay and max attempts.

    Request Syntax:

        [eks-nodegroup-name]:
              aws.eks.nodegroup.present:
              - cluster_name: 'string'
              - node_group_arn: 'string'
              - version: 'string'
              - release_version: 'string'
              - resource_id: 'string'
              - status: 'string'
              - capacity_type: 'ON_DEMAND' | 'SPOT',
              - instance_types:
                - 'string'
              - subnets:
                - 'string'
              - ami_type: 'string'
              - node_role: 'string'
              - disk_size: int
              - scaling_config:
                  desiredSize: int
                  maxSize: int
                  minSize: int
              - update_config:
                  maxUnavailable: int
              - tags:
                  - 'string': 'string'
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            nodes1:
              aws.eks.nodegroup.present:
              - cluster_name: idem-test-cluster
              - node_group_arn: arn:aws:eks:us-west-2:000000000000:nodegroup/prod/idem-test-instance-e1a573b9-4e74-48f0-85c4-c214a8ec8ada/id123
              - version: '1.21'
              - release_version: 1.21.5-20220123
              - resource_id: 'idem-test-cluster-node-group'
              - status: ACTIVE
              - capacity_type: ON_DEMAND
              - instance_types:
                - t1.micro
              - subnets:
                - subnet-31813031
              - ami_type: AL2_x86_64
              - node_role: arn:aws:iam::000000000000:role/idem-test-role-2075a427-24c2-4021-abc3-9b542834addb
              - disk_size: 20
              - scaling_config:
                  desiredSize: 2
                  maxSize: 2
                  minSize: 2
              - update_config:
                  maxUnavailable: 1
              - tags:
                  - Name: idem-test-instance-e1a573b9-4e74-48f0-85c4-c214a8ec8ada



    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.eks.describe_nodegroup(
            ctx, clusterName=cluster_name, nodegroupName=resource_id
        )
        if not before["result"]:
            if not ("ResourceNotFoundException" in before["comment"][0]):
                result["result"] = False
                result["comment"] = before["comment"]
                return result

    if before and before["ret"]:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
                raw_resource=before["ret"]["nodegroup"], idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])
            client_request_token = result["old_state"].get("client_request_token", None)

            # update nodegroup
            update_ret = await hub.exec.aws.eks.nodegroup.update_nodegroup(
                ctx=ctx,
                name=name,
                before=result["old_state"],
                version=version,
                release_version=release_version,
                launch_template=launch_template,
                labels=labels,
                taints=taints,
                scaling_config=scaling_config,
                update_config=update_config,
                client_request_token=client_request_token,
                timeout=timeout,
            )
            result["comment"] = update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = bool(update_ret["ret"])

            if update_ret["ret"] and ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update aws.eks.nodegroup '{name}'",
                )
                for nodegroup_param in update_ret["ret"]:
                    plan_state[nodegroup_param] = update_ret["ret"][nodegroup_param]

            # update tags
            update_tags_ret = await hub.exec.aws.eks.tag.update_eks_tags(
                ctx=ctx,
                resource_arn=result["old_state"].get("nodegroup_arn"),
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            result["result"] = result["result"] and update_tags_ret["result"]
            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if ctx.get("test", False) and update_tags_ret["ret"]:
                plan_state["tags"] = update_tags_ret["ret"].get("tags")
        except Exception as e1:
            result["comment"] = result["comment"] + (f"{e1.__class__.__name__}: {e1}",)
            result["result"] = False
            return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "cluster_name": cluster_name,
                    "subnets": subnets,
                    "node_role": node_role,
                    "scaling_config": scaling_config,
                    "disk_size": disk_size,
                    "instance_types": instance_types,
                    "ami_type": ami_type,
                    "remote_access": remote_access,
                    "labels": labels,
                    "taints": taints,
                    "tags": tags,
                    "client_request_token": client_request_token,
                    "launch_template": launch_template,
                    "update_config": update_config,
                    "capacity_type": capacity_type,
                    "version": version,
                    "release_version": release_version,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.eks.nodegroup", name=name
            )
            return result
        try:
            ret = await hub.exec.boto3.client.eks.create_nodegroup(
                ctx,
                clusterName=cluster_name,
                nodegroupName=name,
                scalingConfig=scaling_config,
                diskSize=disk_size,
                subnets=subnets,
                instanceTypes=instance_types,
                amiType=ami_type,
                remoteAccess=remote_access,
                nodeRole=node_role,
                labels=labels,
                taints=taints,
                tags=tags,
                clientRequestToken=client_request_token,
                launchTemplate=launch_template,
                updateConfig=update_config,
                capacityType=capacity_type,
                version=version,
                releaseVersion=release_version,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result

            resource_id = ret["ret"]["nodegroup"]["nodegroupName"]
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=60,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            hub.log.debug(f"Waiting on creating aws.eks.nodegroup '{name}'")
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "eks",
                    "nodegroup_active",
                    nodegroupName=resource_id,
                    clusterName=cluster_name,
                    WaiterConfig=waiter_config,
                )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.eks.nodegroup", name=name
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.eks.describe_nodegroup(
                ctx, clusterName=cluster_name, nodegroupName=resource_id
            )
            result[
                "new_state"
            ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
                raw_resource=after["ret"]["nodegroup"], idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    cluster_name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes an Amazon EKS node group for a cluster.

    Args:
        name(Text): An Idem name of EKS Node group.
        cluster_name(Text): The name of the cluster to create the node group in.
        resource_id(Text, optional): AWS EKS Node group name. Idem automatically considers this resource being absent
         if this field is not specified.
        timeout(Dict, optional): Timeout configuration for creating or updating cluster.
            * delete (Dict) -- Timeout configuration for deleting cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.
    Request Syntax:
        [nodegroup-name]:
              aws.eks.nodegroup.absent:
                - cluster_name: 'string'
                - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            nodes1:
              aws.eks.nodegroup.absent:
                - cluster_name: idem-test-cluster
                - resource_id: idem-test-nodegroup
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.eks.nodegroup", name=name
        )
        return result
    before = await hub.exec.boto3.client.eks.describe_nodegroup(
        ctx, clusterName=cluster_name, nodegroupName=resource_id
    )
    if not before["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.eks.nodegroup", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
            raw_resource=before["ret"]["nodegroup"], idem_resource_name=name
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.eks.nodegroup", name=name
        )
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
            raw_resource=before["ret"]["nodegroup"], idem_resource_name=name
        )
        try:
            ret = await hub.exec.boto3.client.eks.delete_nodegroup(
                ctx, clusterName=cluster_name, nodegroupName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            else:
                waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                    default_delay=60,
                    default_max_attempts=40,
                    timeout_config=timeout.get("delete") if timeout else None,
                )
                try:
                    await hub.tool.boto3.client.wait(
                        ctx,
                        "eks",
                        "nodegroup_deleted",
                        nodegroupName=resource_id,
                        clusterName=cluster_name,
                        WaiterConfig=waiter_config,
                    )
                except Exception as e:
                    result["comment"] = result["comment"] + (str(e),)
                    result["result"] = False
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.eks.nodegroup", name=name
            )
        except Exception as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the Amazon EKS managed node groups associated with the specified cluster in your Amazon Web Services
    account in the specified Region. Self-managed node groups are not listed.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.eks.nodegroup
    """

    result = {}
    cluster_ret = await hub.exec.boto3.client.eks.list_clusters(ctx)
    if not cluster_ret["result"]:
        hub.log.debug(f"Could not describe cluster {cluster_ret['comment']}")
        return {}
    cluster_node_groups = []
    try:
        for name in cluster_ret["ret"]["clusters"]:
            node_group_ret = await hub.exec.boto3.client.eks.list_nodegroups(
                ctx, clusterName=name
            )
            if not node_group_ret["result"]:
                result["comment"] = node_group_ret["comment"]
                return result
            for group in node_group_ret["ret"]["nodegroups"]:
                cluster_node_groups.append({"name": name, "group": group})
    except Exception as e:
        result["comment"] = str(e)
        result["result"] = False
        return result

    for node_groups in cluster_node_groups:
        cluster_name = node_groups["name"]
        node_group = node_groups["group"]
        describe_ret = await hub.exec.boto3.client.eks.describe_nodegroup(
            ctx, clusterName=cluster_name, nodegroupName=node_group
        )
        nodegroup = describe_ret["ret"]["nodegroup"]
        resource_id = f"{nodegroup['clusterName']}-{nodegroup['nodegroupName']}"
        resource_translated = (
            hub.tool.aws.eks.conversion_utils.convert_raw_nodegroup_to_present(
                raw_resource=nodegroup, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.eks.nodegroup.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result
